# -*- coding: utf-8 -*-
#
# Copyright 2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Test the proxy UI."""

import qt4reactor

from PyQt4.QtCore import QObject, pyqtSignal
from twisted.internet import defer, reactor
from twisted.trial.unittest import TestCase

from ubuntu_kylin_sso.qt import proxy_dialog

# pylint: disable=E1101,W0212,


class FakeSignal(object):
    """A fake qt signal."""

    def __init__(self):
        """Create a new instance."""
        self.called = []

    def connect(self, cb):
        """Connect signals."""
        self.called.append(('connect', cb))


class FakeWidget(object):
    """A fake widget that contains text."""

    def __init__(self, object_name, called, text=''):
        """Create a new instance."""
        self.object_name = object_name
        self.called = called
        self.internal_text = text
        self.clicked = FakeSignal()

    # pylint: disable=C0103
    def setText(self, text):
        """Set the text of the widget."""
        self.internal_text = text
        if self.object_name not in self.called:
            self.called[self.object_name] = [('setText', text)]
        else:
            self.called[self.object_name].append(('setText', text))

    def text(self):
        """Return the text."""
        if self.object_name not in self.called:
            self.called[self.object_name] = [('text',)]
        else:
            self.called[self.object_name].append(('text',))
        return self.internal_text

    def setPixmap(self, pixmap):
        """Set a pixmap."""
        if self.object_name not in self.called:
            self.called[self.object_name] = [('setPixmap', pixmap)]
        else:
            self.called[self.object_name].append(('setPixmap', pixmap))

    def setVisible(self, visible):
        """Set the ui element visible."""
        if self.object_name not in self.called:
            self.called[self.object_name] = [('setVisible', visible)]
        else:
            self.called[self.object_name].append(('setVisible', visible))

    # pylint: enable=C0103


class FakeUIProxyCredsDialog(object):
    """The fake .ui for the creds dialog."""

    ui_labels = ('title_label', 'explanation_label', 'connection_label',
                   'error_label', 'username_label', 'password_label',
                   'domain_label', 'logo_label')
    ui_buttons = ('help_button', 'cancel_button', 'save_button')
    ui_entries = ('username_entry', 'password_entry')

    def __init__(self):
        """Create a new instance."""
        self.called = {}
        for name in self.ui_labels + self.ui_buttons + self.ui_entries:
            setattr(self, name, FakeWidget(name, self.called))

    # pylint: disable=C0103
    def setupUi(self, dialog):
        """Set the ui."""
        self.called['FakeUIProxyCredsDialog'] = ('setupUi', dialog)
    # pylint: enable=C0103


class FakeQIcon(object):
    """QIcon class."""

    def __init__(self):
        """Create a new instance."""
        self.called = []

    # pylint: disable=C0103
    def fromTheme(self, icon_id):
        """Return the icon with the given id."""
        self.called.append(('fromTheme', icon_id))
        return self
    # pylint: enable=C0103

    # pylint: disable=C0103
    def pixmap(self, h, w):
        """Get the pixmap with the given size."""
        self.called.append(('pixmap', h, w))
        return self
    # pylint: disable=C0103


class ProxyCredsDialogTest(TestCase):
    """Test the proxy creds dialog."""

    @defer.inlineCallbacks
    def setUp(self):
        """Set tests."""
        yield super(ProxyCredsDialogTest, self).setUp()
        self.patch(proxy_dialog, 'Ui_ProxyCredsDialog',
                   FakeUIProxyCredsDialog)
        self.icon = FakeQIcon()
        self.patch(proxy_dialog, 'QIcon', self.icon)

    def test_set_labels(self):
        """Ensure the correct labels are used."""
        dialog = proxy_dialog.ProxyCredsDialog()
        for label in dialog.ui.ui_labels:
            self.assertIn(label, dialog.ui.called, label)

    def test_set_buttons(self):
        """Ensure that buttons are set."""
        dialog = proxy_dialog.ProxyCredsDialog()
        for label in dialog.ui.ui_buttons:
            self.assertIn(label, dialog.ui.called)
        # assert that the clicked was set
        self.assertIn(('connect', dialog._on_save_clicked),
                      dialog.ui.save_button.clicked.called)
        self.assertIn(('connect', dialog._on_cancel_clicked),
                      dialog.ui.cancel_button.clicked.called)

    def test_domain_empty(self):
        """Test that if the domain is not passed we set ''"""
        dialog = proxy_dialog.ProxyCredsDialog()
        self.assertIn('domain_label', dialog.ui.called)
        self.assertIn(('setText', ''), dialog.ui.called['domain_label'])

    def test_domain_not_empty(self):
        """Test that we do set the domain."""
        domain = '192.168.1.100'
        dialog = proxy_dialog.ProxyCredsDialog(domain=domain)
        self.assertIn('domain_label', dialog.ui.called)
        self.assertIn(('setText', domain), dialog.ui.called['domain_label'])

    @defer.inlineCallbacks
    def test_retry_present_creds(self):
        """Test that we do set the error label correctly."""
        creds = dict(username='username', password='password')
        called = []
        dialog = proxy_dialog.ProxyCredsDialog(retry=True)

        def fake_get_credentials(app_name):
            """Fake the get creds."""
            called.append(('get_credentials', app_name))
            return defer.succeed(creds)

        self.patch(dialog.keyring, 'get_credentials', fake_get_credentials)

        yield dialog._load_creds()
        self.assertIn('get_credentials', called[0])
        self.assertIn('error_label', dialog.ui.called)
        self.assertIn(('setVisible', True), dialog.ui.called['error_label'])
        self.assertIn(('setText', creds['username']),
                      dialog.ui.called['username_entry'])
        self.assertIn(('setText', creds['password']),
                      dialog.ui.called['password_entry'])

    @defer.inlineCallbacks
    def test_retry_none_creds(self):
        """Test that we do set the error label correctly."""
        called = []
        dialog = proxy_dialog.ProxyCredsDialog(retry=True)

        def fake_get_credentials(app_name):
            """Fake the get creds."""
            called.append(('get_credentials', app_name))
            return defer.succeed(None)

        self.patch(dialog.keyring, 'get_credentials', fake_get_credentials)

        yield dialog._load_creds()
        self.assertIn('get_credentials', called[0])
        self.assertIn('error_label', dialog.ui.called)
        self.assertIn(('setVisible', True), dialog.ui.called['error_label'])
        self.assertNotIn('username_entry', dialog.ui.called)
        self.assertNotIn('password_entry', dialog.ui.called)

    @defer.inlineCallbacks
    def test_retry_exception_creds(self):
        """Test that we do set the error label correctly."""
        called = []
        dialog = proxy_dialog.ProxyCredsDialog(retry=True)

        def fake_get_credentials(app_name):
            """Fake the get creds."""
            called.append(('get_credentials', app_name))
            return defer.fail(Exception('Fail test'))

        self.patch(dialog.keyring, 'get_credentials', fake_get_credentials)

        yield dialog._load_creds()
        self.assertIn('get_credentials', called[0])
        self.assertIn('error_label', dialog.ui.called)
        self.assertIn(('setVisible', True), dialog.ui.called['error_label'])
        self.assertNotIn('username_entry', dialog.ui.called)
        self.assertNotIn('password_entry', dialog.ui.called)

    def test_is_not_retry(self):
        """Test that we do set the error label correctly."""
        dialog = proxy_dialog.ProxyCredsDialog(retry=False)
        self.assertIn('error_label', dialog.ui.called)
        self.assertIn(('setVisible', False), dialog.ui.called['error_label'])

    def test_icon(self):
        """Test that we correctly set the icon."""
        dialog = proxy_dialog.ProxyCredsDialog()
        self.assertIn(('fromTheme', 'gtk-dialog-authentication'),
                      self.icon.called)
        self.assertIn(('pixmap', 48, 48),
                      self.icon.called)
        self.assertIn(('setPixmap', self.icon),
                      dialog.ui.called['logo_label'])

    def test_on_cancel_clicked(self):
        """Test the on cancel clicked event."""
        dialog = proxy_dialog.ProxyCredsDialog()
        called = []

        def fake_done(result):
            """Fake the done method."""
            called.append(('done', result))

        self.patch(dialog, 'done', fake_done)

        dialog._on_cancel_clicked()
        self.assertIn(('done', proxy_dialog.USER_CANCELLATION), called)

    def assert_save_button(self, set_creds_callback, result_number):
        """Test the save button execution."""
        dialog = proxy_dialog.ProxyCredsDialog()
        called = []

        def fake_done(result):
            """Fake the done method."""
            called.append(('done', result))

        self.patch(dialog, 'done', fake_done)

        def fake_set_credentials(app_name, creds):
            """Fake an error."""
            called.append(('set_credentials', app_name, creds))
            return set_creds_callback()

        self.patch(dialog.keyring, 'set_credentials', fake_set_credentials)

        dialog._on_save_clicked()
        self.assertIn(('done', result_number), called)

    def test_on_save_clicked_exception(self):
        """Test that the exception is handled correctly."""
        set_creds_cb = lambda: defer.fail(Exception())
        result_number = proxy_dialog.EXCEPTION_RAISED
        self.assert_save_button(set_creds_cb, result_number)

    def test_on_save_clicked_correct(self):
        """Test that we do save the creds."""
        set_creds_cb = lambda: defer.succeed(True)
        result_number = proxy_dialog.USER_SUCCESS
        self.assert_save_button(set_creds_cb, result_number)


class FakeArgs(object):
    """Some fake args."""

    def __init__(self, domain='', retry=False):
        """Create a new instance."""
        self.domain = domain
        self.retry = retry


class FakeCredsDialog(QObject):
    """A fake dialog."""

    finished = pyqtSignal(int)

    def __init__(self, called, code):
        """Create a new instance."""
        super(FakeCredsDialog, self).__init__()
        self.domain = None
        self.retry = None
        self.called = called
        self.return_code = code

    def __call__(self, domain, retry):
        """Fake init."""
        self.domain = domain
        self.retry = retry
        return self

    def show(self):
        """Show the dialog."""
        self.called.append(('show',))

    def exec_(self):
        """Show the dialog."""
        self.called.append(('exec_',))
        return self.return_code


class FakeQApplication(object):
    """A fake QApplication."""

    test_instance = None

    def __init__(self, called):
        """Create a new instance."""
        self.called = called
        self.args = None
        self.code = None
        FakeQApplication.test_instance = self

    def __call__(self, args):
        """Fake construntor."""
        self.args = args

    @classmethod
    def instance(cls):
        """Return the instance."""
        return FakeQApplication.test_instance

    def exit(self, code):
        """Exit the app."""
        self.called.append(('exit', code))
        self.code = code


class MainTest(TestCase):
    """Test the main method."""

    @defer.inlineCallbacks
    def setUp(self):
        """Set the different tests."""
        yield super(MainTest, self).setUp()
        self.called = []

        self.return_code = 0
        self.dialog = FakeCredsDialog(self.called, self.return_code)
        self.patch(proxy_dialog, 'ProxyCredsDialog', self.dialog)

        self.args = FakeArgs(domain='domain', retry=False)
        self.patch(proxy_dialog, 'parse_args', lambda: self.args)

        self.exit_code = None

        self.app = FakeQApplication(self.called)

        self.patch(proxy_dialog, 'QApplication', self.app)

    def _set_old_platform(self, platform):
        """Set back the platform."""
        proxy_dialog.sys.platform = platform

    def _assert_basic_start(self):
        """Assert the common code is actually ran."""
        self.assertEqual(self.args.domain, self.dialog.domain)
        self.assertEqual(self.args.retry, self.dialog.retry)

    def test_linux_main(self):
        """Test the main method on linux."""
        old_platform = proxy_dialog.sys.platform
        proxy_dialog.sys.platform = 'linux'
        self.addCleanup(self._set_old_platform, old_platform)

        def fake_exit(code):
            """Fake sys exit."""
            self.called.append(('exit', code))

        self.patch(proxy_dialog.sys, 'exit', fake_exit)

        proxy_dialog.main()

        self._assert_basic_start()
        self.assertIn(('exec_',), self.called)
        self.assertIn(('exit', self.return_code), self.called)

    def test_windows_main(self):
        """Test the main method on windows."""
        old_platform = proxy_dialog.sys.platform
        proxy_dialog.sys.platform = 'win32'
        self.addCleanup(self._set_old_platform, old_platform)

        def fake_install():
            """Fake the reactor installation."""
            self.called.append(('install',))

        self.patch(qt4reactor, 'install', fake_install)

        def fake_start():
            """Fake reactor.start."""
            self.called.append(('run',))

        self.patch(reactor, 'run', fake_start)

        def fake_exit(code):
            """Fake the exit method."""
            self.called.append(('exit_code', code))

        self.patch(proxy_dialog, 'exit_code', fake_exit)

        proxy_dialog.main()

        self._assert_basic_start()
        self.assertIn(('show',), self.called)
        self.assertIn(('run',), self.called)

        self.dialog.finished.emit(self.return_code)
        self.assertIn(('exit_code', self.return_code), self.called)

    def test_exit_code_windows(self):
        """"Test the exit code function."""
        old_platform = proxy_dialog.sys.platform
        proxy_dialog.sys.platform = 'win32'
        self.addCleanup(self._set_old_platform, old_platform)

        def fake_stop():
            """Fake reactor.start."""
            self.called.append(('stop',))

        self.patch(reactor, 'stop', fake_stop)

        proxy_dialog.exit_code(self.return_code)
        self.assertIn(('stop',), self.called)
        self.assertIn(('exit', self.return_code), self.called)



# -*- coding: utf-8 -*-
#
# Copyright 2011-2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Tests for the timestamp sync classes."""

from twisted.internet import defer
from twisted.trial.unittest import TestCase
from twisted.web import resource

from ubuntuone.devtools.testing.txwebserver import HTTPWebServer

from ubuntu_kylin_sso.utils.webclient import timestamp, webclient_module


class FakedError(Exception):
    """Stub to replace Request.error."""


class RootResource(resource.Resource):
    """A root resource that logs the number of calls."""

    isLeaf = True

    def __init__(self, *args, **kwargs):
        """Initialize this fake instance."""
        resource.Resource.__init__(self, *args, **kwargs)
        self.count = 0
        self.request_headers = []

    # pylint: disable=C0103
    def render_HEAD(self, request):
        """Increase the counter on each render."""
        self.request_headers.append(request.requestHeaders)
        self.count += 1
        return ""


class MockWebServer(HTTPWebServer):
    """A mock webserver for testing."""

    def __init__(self):
        """Create a new server."""
        super(MockWebServer, self).__init__(RootResource())


class TimestampCheckerTestCase(TestCase):
    """Tests for the timestamp checker."""

    timeout = 5

    @defer.inlineCallbacks
    def setUp(self):
        yield super(TimestampCheckerTestCase, self).setUp()
        self.ws = MockWebServer()
        self.ws.start()
        self.addCleanup(self.ws.stop)
        self.webclient_class = webclient_module().WebClient
        self.patch(timestamp.TimestampChecker, "SERVER_IRI", self.ws.get_iri())

    @defer.inlineCallbacks
    def test_returned_value_is_int(self):
        """The returned value is an integer."""
        checker = timestamp.TimestampChecker(self.webclient_class)
        result = yield checker.get_faithful_time()
        self.assertEqual(type(result), int)

    @defer.inlineCallbacks
    def test_first_call_does_head(self):
        """The first call gets the clock from our web."""
        checker = timestamp.TimestampChecker(self.webclient_class)
        yield checker.get_faithful_time()
        self.assertEqual(self.ws.root.count, 1)

    @defer.inlineCallbacks
    def test_second_call_is_cached(self):
        """For the second call, the time is cached."""
        checker = timestamp.TimestampChecker(self.webclient_class)
        yield checker.get_faithful_time()
        yield checker.get_faithful_time()
        self.assertEqual(self.ws.root.count, 1)

    @defer.inlineCallbacks
    def test_after_timeout_cache_expires(self):
        """After some time, the cache expires."""
        fake_timestamp = 1
        self.patch(timestamp.time, "time", lambda: fake_timestamp)
        checker = timestamp.TimestampChecker(self.webclient_class)
        yield checker.get_faithful_time()
        fake_timestamp += timestamp.TimestampChecker.CHECKING_INTERVAL
        yield checker.get_faithful_time()
        self.assertEqual(self.ws.root.count, 2)

    @defer.inlineCallbacks
    def test_server_error_means_skew_not_updated(self):
        """When server can't be reached, the skew is not updated."""
        fake_timestamp = 1
        self.patch(timestamp.time, "time", lambda: fake_timestamp)
        checker = timestamp.TimestampChecker(self.webclient_class)
        failing_get_server_time = lambda _: defer.fail(FakedError())
        self.patch(checker, "get_server_time", failing_get_server_time)
        yield checker.get_faithful_time()
        self.assertEqual(checker.skew, 0)
        self.assertEqual(checker.next_check,
                    fake_timestamp + timestamp.TimestampChecker.ERROR_INTERVAL)

    @defer.inlineCallbacks
    def test_server_date_sends_nocache_headers(self):
        """Getting the server date sends the no-cache headers."""
        checker = timestamp.TimestampChecker(self.webclient_class)
        yield checker.get_server_date_header(self.ws.get_iri())
        self.assertEqual(len(self.ws.root.request_headers), 1)
        headers = self.ws.root.request_headers[0]
        result = headers.getRawHeaders("Cache-Control")
        self.assertEqual(result, ["no-cache"])



# -*- coding: utf-8 -*-
#
# Copyright 2011-2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Tests for the network manager."""

from ctypes import windll
from mocker import MockerTestCase
from twisted.internet.defer import inlineCallbacks

from ubuntu_kylin_sso.tests import TestCase
from ubuntu_kylin_sso.networkstate import NetworkFailException
from ubuntu_kylin_sso.networkstate.windows import (
    is_machine_connected,
    NetworkManager,
    NetworkManagerState,
    ONLINE,
    OFFLINE)


class TestNetworkManager(MockerTestCase):
    """Test he Network Manager."""

    def setUp(self):
        super(TestNetworkManager, self).setUp()
        self.connection_info = self.mocker.mock()
        self.connection_no_info = self.mocker.mock()
        self.disconnected = self.mocker.mock()
        self.manager = NetworkManager(self.connection_no_info,
                                      self.connection_info, self.disconnected)

    def test_connection_made(self):
        """Ensure db is called."""
        self.connection_info()
        self.mocker.replay()
        self.manager.ConnectionMade()

    def test_connection_made_no_cb(self):
        """Ensure db is called."""
        self.manager.connected_cb_info = None
        self.mocker.replay()
        self.manager.ConnectionMade()

    def test_connection_made_no_info(self):
        """Ensure db is called."""
        self.connection_no_info()
        self.mocker.replay()
        self.manager.ConnectionMadeNoQOCInfo()

    def test_connection_made_no_info_no_cb(self):
        """Ensure db is called."""
        self.manager.connected_cb = None
        self.mocker.replay()
        self.manager.ConnectionMadeNoQOCInfo()

    def test_disconnection(self):
        """Ensure db is called."""
        self.disconnected()
        self.mocker.replay()
        self.manager.ConnectionLost()

    def test_disconnection_no_cb(self):
        """Ensure db is called."""
        self.manager.disconnected_cb = None
        self.mocker.replay()
        self.manager.ConnectionLost()


class TestNetworkManagerState(MockerTestCase):
    """Test the Network Manager State."""

    def setUp(self):
        super(TestNetworkManagerState, self).setUp()
        self.network_manager = self.mocker.mock()
        self.is_connected = self.mocker.replace(
            'ubuntu_kylin_sso.networkstate.windows.is_machine_connected')
        self.thread = self.mocker.mock()
        self.cb = self.mocker.mock()
        self.state = NetworkManagerState(self.cb)

    def test_connection_made(self):
        """Test that the cb is actually called."""
        self.cb(ONLINE)
        self.mocker.replay()
        self.state.connection_made()

    def test_connection_lost(self):
        """Test that the cb is actually called."""
        self.cb(OFFLINE)
        self.mocker.replay()
        self.state.connection_lost()

    def test_find_online_state_not_connected(self):
        """Test that we do find the online state correctly."""
        self.is_connected()
        self.mocker.result(False)
        self.cb(OFFLINE)
        self.mocker.result(self.thread)
        self.thread.daemon = True
        self.thread.start()
        self.mocker.replay()
        self.state.find_online_state(listener=self.network_manager,
                                     listener_thread=self.thread)

    def test_find_online_state_connected(self):
        """Test that we do find the online state correctly."""
        self.is_connected()
        self.mocker.result(ONLINE)
        self.cb(ONLINE)
        self.mocker.result(self.thread)
        self.thread.daemon = True
        self.thread.start()
        self.mocker.replay()
        self.state.find_online_state(listener=self.network_manager,
                                     listener_thread=self.thread)


class FakeWininet(object):
    """Fake wininet for windll."""

    connection_state = -1

    # pylint: disable=C0103
    def InternetGetConnectedState(self, *args, **kwargs):
        """Fake InternetGetConnectedState function from wininet."""
        return self.connection_state
    # pylint: enable=C0103


class FakeWininetException(object):
    """Fake wininet for windll."""

    connection_state = -1

    # pylint: disable=C0103
    def InternetGetConnectedState(self, *args, **kwargs):
        """Fake InternetGetConnectedState function from wininet."""
        raise Exception()
    # pylint: enable=C0103


class TestConnection(TestCase):
    """Test the state of the connection."""

    @inlineCallbacks
    def setUp(self):
        """Setup the mocker dbus object tree."""
        yield super(TestConnection, self).setUp()
        self.patch(windll, "wininet", FakeWininet())

    @inlineCallbacks
    def test_is_machine_connected_connected(self):
        """Fake the NetworkManagerState."""
        self.patch(FakeWininet, "connection_state", 1)
        result = yield is_machine_connected()
        self.assertTrue(result)

    @inlineCallbacks
    def test_is_machine_connected_disconnected(self):
        """Fake the NetworkManagerState."""
        self.patch(FakeWininet, "connection_state", 0)
        result = yield is_machine_connected()
        self.assertFalse(result)

    @inlineCallbacks
    def test_is_machine_connected_error(self):
        """Fake the NetworkManagerState."""
        self.patch(windll, "wininet", FakeWininetException())
        yield self.assertFailure(is_machine_connected(), NetworkFailException)



# -*- coding: utf-8 -*-
#
# Copyright 2014 National University of Defense Technology(NUDT) & Kylin Ltd
# Copyright 2012-2013 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Qt main loop runner."""

# pylint: disable=E0611,F0401

import sys

# pylint: disable=W0621
import dbus.mainloop.qt
# pylint: enable=W0621

from PyQt4 import QtCore

from ubuntu_kylin_sso.utils.locale import fix_turkish_locale


TIMERS = set()


def timeout_func(interval, callback, *a, **kw):
    """Delay import of dynamic bindings to avoid crashes."""

    # QTimers don't support priorities
    kw.pop("priority", None)

    timer = QtCore.QTimer()
    TIMERS.add(timer)

    def _callback():
        """Call the real callback, with arguments, until it returns False."""
        if timer in TIMERS:
            result = callback(*a, **kw)
            if not result:
                timer.stop()
                TIMERS.remove(timer)
                # Probably overkill
                timer.deleteLater()

    timer.timeout.connect(_callback)
    timer.start(interval)


def shutdown_func(*a, **kw):
    """Delay import of dynamic bindings to avoid crashes."""
    QtCore.QCoreApplication.instance().exit()


def start_setup():
    """Setup the env to run the service."""
    fix_turkish_locale()
    # this has to be created before calling dbus.mainloop.qt.DBusQtMainLoop
    loop = QtCore.QCoreApplication(sys.argv)
    dbus.mainloop.qt.DBusQtMainLoop(set_as_default=True)
    return loop


def run_func(loop):
    """Delay import of dynamic bindings to avoid crashes."""
    loop.exec_()



# -*- coding: utf-8 -*-
#
# test_keyring - tests for ubuntu_kylin_sso.keyring
#
# Author: Alejandro J. Cura <alecu@canonical.com>
# Author: Natalia B. Bidart <natalia.bidart@canonical.com>
#
# Copyright 2010-2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Tests for the keyring common module."""

from __future__ import unicode_literals

from twisted.trial.unittest import TestCase

from ubuntu_kylin_sso import keyring


class TestGetHostname(TestCase):
    """Test the function that gets the hostname."""

    def test_get_hostname(self):
        """The common case."""
        fake_hostname = "fake hostname"
        self.patch(keyring.socket, "gethostname", lambda: fake_hostname)
        self.assertEqual(keyring.gethostname(), fake_hostname)

    def test_get_hostname_uses_filesystem_encoding(self):
        """The fs encoding is used to decode the name returned by socket."""
        fake_hostname = "Привет-ПК"
        hostname_koi8r = fake_hostname.encode("koi8-r")
        self.patch(keyring.socket, "gethostname", lambda: hostname_koi8r)
        self.patch(keyring.sys, "getfilesystemencoding", lambda: "koi8-r")
        self.assertEqual(keyring.gethostname(), fake_hostname)


class TestTokenNameBuilder(TestCase):
    """Test the function that builds the token name."""

    def check_build(self, sample_app_name, sample_hostname, expected_result):
        """Check the build of a given token."""
        self.patch(keyring, "gethostname", lambda *a: sample_hostname)
        result = keyring.get_token_name(sample_app_name)
        self.assertEqual(result, expected_result)

    def test_get_simple_token_name(self):
        """A simple token name is built right."""
        sample_app_name = "UbuntuTwo"
        sample_hostname = "Darkstar"
        expected_result = "UbuntuTwo @ Darkstar"
        self.check_build(sample_app_name, sample_hostname, expected_result)

    def test_get_complex_token_name_for_app_name(self):
        """A complex token name is built right too."""
        sample_app_name = "Ubuntu @ Eleven"
        sample_hostname = "Mate+Cocido"
        expected_result = "Ubuntu @ Eleven @ Mate+Cocido"
        self.check_build(sample_app_name, sample_hostname, expected_result)

    def test_get_complex_token_name_for_hostname(self):
        """A complex token name is built right too."""
        sample_app_name = "Ubuntu Eleven"
        sample_hostname = "Mate @ Cocido"
        expected_result = "Ubuntu Eleven @ Mate AT Cocido"
        self.check_build(sample_app_name, sample_hostname, expected_result)

    def test_get_unicode_appname_token_name(self):
        """A token name with unicode in the app name."""
        sample_app_name = "Ubuntu 四百六十九"
        sample_hostname = "Darkstar"
        expected_result = "Ubuntu 四百六十九 @ Darkstar"
        self.check_build(sample_app_name, sample_hostname, expected_result)

    def test_get_utf8_hostname_token_name(self):
        """A token name with utf8 in the host name."""
        sample_app_name = "Ubuntu Eleven"
        sample_hostname = "Привет-ПК"
        expected_result = "Ubuntu Eleven @ Привет-ПК"
        self.check_build(sample_app_name, sample_hostname, expected_result)



# -*- coding: utf-8 -*-
#
# Copyright 2010-2013 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Single Sign On account management.

All the methods in Account expect unicode as parameters.

"""

from __future__ import unicode_literals

import os
import re

from twisted.internet import defer

from ubuntu_kylin_sso import SSO_AUTH_BASE_URL
from ubuntu_kylin_sso.logger import setup_logging
from ubuntu_kylin_sso.utils import compat, webclient
from ubuntu_kylin_sso.utils.webclient import restful
from ubuntu_kylin_sso.utils.webclient.common import WebClientError


logger = setup_logging("ubuntu_kylin_sso.account")
SERVICE_URL = "%s/api/1.0/" % SSO_AUTH_BASE_URL
SSO_STATUS_OK = 'ok'
SSO_STATUS_ERROR = 'error'


class InvalidEmailError(Exception):
    """The email is not valid."""


class InvalidPasswordError(Exception):
    """The password is not valid.

    Must provide at least 8 characters, one upper case, one number.
    """


class RegistrationError(Exception):
    """The registration failed."""


class AuthenticationError(Exception):
    """The authentication failed."""


class EmailTokenError(Exception):
    """The email token is not valid."""


class ResetPasswordTokenError(Exception):
    """The token for password reset could not be generated."""


class NewPasswordError(Exception):
    """The new password could not be set."""


class Account(object):
    """Login and register users using the Ubuntu Single Sign On service."""

    def __init__(self, service_url=None):
        """Create a new SSO Account manager."""
        if service_url is not None:
            self.service_url = service_url
        else:
            self.service_url = os.environ.get('USSOC_SERVICE_URL', SERVICE_URL)
        logger.info('Creating a new SSO access layer for service url %r',
                     self.service_url)
        assert self.service_url.endswith("/")

    def _valid_email(self, email):
        """Validate the given email."""
        return email is not None and '@' in email

    def _valid_password(self, password):
        """Validate the given password."""
        # res = (len(password) > 7 and  # at least 8 characters
        #        re.search('[A-Z]', password) and  # one upper case
        #        re.search('\d+', password))  # one number
        # weaken the password requirement
        res = (len(password) > 7)  # one number
        return res

    def _format_webservice_errors(self, errdict):
        """Turn each list of strings in the errdict into a LF separated str."""
        result = {}
        for key, val in errdict.items():
            # workaround until bug #624955 is solved
            if isinstance(val, compat.basestring):
                result[key] = val
            else:
                result[key] = "\n".join(val)
        return result

    @defer.inlineCallbacks
    def generate_captcha(self, filename):
        """Generate a captcha using the SSO service."""
        logger.debug('generate_captcha: requesting captcha, filename: %r',
                     filename)
        restful_client = restful.RestfulClient(self.service_url)
        try:
            captcha = yield restful_client.restcall("captchas.new")
        finally:
            restful_client.shutdown()

        # download captcha and save to 'filename'
        logger.debug('generate_captcha: server answered: %r', captcha)
        wc = webclient.webclient_factory()
        try:
            response = yield wc.request(captcha['image_url'])
            with open(filename, 'wb') as f:
                f.write(response.content)
        except:
            msg = 'generate_captcha crashed while downloading the image.'
            logger.exception(msg)
            raise
        finally:
            wc.shutdown()

        defer.returnValue(captcha['captcha_id'])

    @defer.inlineCallbacks
    def register_user(self, email, password, displayname,
                      captcha_id, captcha_solution):
        """Register a new user with 'email' and 'password'."""
        logger.debug('register_user: email: %r password: <hidden>, '
                     'displayname: %r, captcha_id: %r, captcha_solution: %r',
                     email, displayname, captcha_id, captcha_solution)
        restful_client = restful.RestfulClient(self.service_url)
        try:
            if not self._valid_email(email):
                logger.error('register_user: InvalidEmailError for email: %r',
                             email)
                raise InvalidEmailError()
            if not self._valid_password(password):
                logger.error('register_user: InvalidPasswordError')
                raise InvalidPasswordError()

            result = yield restful_client.restcall("registration.register",
                        email=email, password=password,
                        displayname=displayname,
                        captcha_id=captcha_id,
                        captcha_solution=captcha_solution)
        finally:
            restful_client.shutdown()
        logger.info('register_user: email: %r result: %r', email, result)

        if result['status'].lower() == SSO_STATUS_ERROR:
            errorsdict = self._format_webservice_errors(result['errors'])
            raise RegistrationError(errorsdict)
        elif result['status'].lower() != SSO_STATUS_OK:
            raise RegistrationError('Received unknown status: %s' % result)
        else:
            defer.returnValue(email)

    @defer.inlineCallbacks
    def login(self, email, password, token_name):
        """Login a user with 'email' and 'password'."""
        logger.debug('login: email: %r password: <hidden>, token_name: %r',
                     email, token_name)
        restful_client = restful.RestfulClient(self.service_url,
                                               username=email,
                                               password=password)
        try:
            credentials = yield restful_client.restcall(
                        "authentications.authenticate",
                        token_name=token_name)
        except WebClientError:
            logger.exception('login failed with:')
            raise AuthenticationError()
        finally:
            restful_client.shutdown()

        logger.debug('login: authentication successful! consumer_key: %r, '
                     'token_name: %r', credentials['consumer_key'], token_name)
        defer.returnValue(credentials)

    @defer.inlineCallbacks
    def is_validated(self, token):
        """Return if user with 'email' and 'password' is validated."""
        logger.debug('is_validated: requesting accounts.me() info.')
        restful_client = restful.RestfulClient(self.service_url,
                                               oauth_credentials=token)
        try:
            me_info = yield restful_client.restcall("accounts.me")
        finally:
            restful_client.shutdown()
        key = 'preferred_email'
        result = key in me_info and me_info[key] is not None

        logger.info('is_validated: consumer_key: %r, result: %r.',
                    token['consumer_key'], result)
        defer.returnValue(result)

    @defer.inlineCallbacks
    def validate_email(self, email, password, email_token, token_name):
        """Validate an email token for user with 'email' and 'password'."""
        logger.debug('validate_email: email: %r password: <hidden>, '
                     'email_token: %r, token_name: %r.',
                     email, email_token, token_name)
        credentials = yield self.login(email=email, password=password,
                                       token_name=token_name)
        restful_client = restful.RestfulClient(self.service_url,
                                               oauth_credentials=credentials)
        try:
            result = yield restful_client.restcall("accounts.validate_email",
                                                   email_token=email_token)
        finally:
            restful_client.shutdown()
        logger.info('validate_email: email: %r result: %r', email, result)
        if 'errors' in result:
            errorsdict = self._format_webservice_errors(result['errors'])
            raise EmailTokenError(errorsdict)
        elif 'email' in result:
            defer.returnValue(credentials)
        else:
            raise EmailTokenError('Received invalid reply: %s' % result)

    @defer.inlineCallbacks
    def request_password_reset_token(self, email):
        """Request a token to reset the password for the account 'email'."""
        restful_client = restful.RestfulClient(self.service_url)
        try:
            operation = "registration.request_password_reset_token"
            result = yield restful_client.restcall(operation, email=email)
        except WebClientError as e:
            logger.exception('request_password_reset_token failed with:')
            raise ResetPasswordTokenError(e[1].split('\n')[0])
        finally:
            restful_client.shutdown()

        if result['status'].lower() == SSO_STATUS_OK:
            defer.returnValue(email)
        else:
            raise ResetPasswordTokenError('Received invalid reply: %s' %
                                          result)

    @defer.inlineCallbacks
    def set_new_password(self, email, token, new_password):
        """Set a new password for the account 'email' to be 'new_password'.

        The 'token' has to be the one resulting from a call to
        'request_password_reset_token'.

        """
        restful_client = restful.RestfulClient(self.service_url)
        try:
            result = yield restful_client.restcall(
                                        "registration.set_new_password",
                                        email=email, token=token,
                                        new_password=new_password)
        except WebClientError as e:
            logger.exception('set_new_password failed with:')
            raise NewPasswordError(e[1].split('\n')[0])
        finally:
            restful_client.shutdown()

        if result['status'].lower() == SSO_STATUS_OK:
            defer.returnValue(email)
        else:
            raise NewPasswordError('Received invalid reply: %s' % result)



# -*- coding: utf-8 -*-
#
# Copyright 2011-2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Test the gsettings parser."""

import logging

from twisted.trial.unittest import TestCase
from ubuntuone.devtools.handlers import MementoHandler

from ubuntu_kylin_sso.utils.webclient import gsettings
from ubuntu_kylin_sso.utils.webclient.tests import (
    BASE_GSETTINGS_VALUES,
    TEMPLATE_GSETTINGS_OUTPUT,
)


class ProxySettingsTestCase(TestCase):
    """Test the getting of the proxy settings."""

    def test_gsettings_cmdline_correct(self):
        """The command line used to get the proxy settings is the right one."""
        expected = "gsettings list-recursively org.gnome.system.proxy".split()
        called = []

        def append_output(args):
            """Append the output and return some settings."""
            called.append(args)
            return TEMPLATE_GSETTINGS_OUTPUT.format(**BASE_GSETTINGS_VALUES)

        self.patch(gsettings.subprocess, "check_output", append_output)
        gsettings.get_proxy_settings()
        self.assertEqual(called[0], expected)

    def test_gsettings_parser_none(self):
        """Test a parser of gsettings."""
        expected = {}
        fake_output = TEMPLATE_GSETTINGS_OUTPUT.format(**BASE_GSETTINGS_VALUES)
        self.patch(gsettings.subprocess, "check_output",
                   lambda _: fake_output)
        ps = gsettings.get_proxy_settings()
        self.assertEqual(ps, expected)

    def _assert_parser_anonymous(self, scheme):
        """Assert the parsing of anonymous settings."""
        template_values = dict(BASE_GSETTINGS_VALUES)
        expected_host = "expected_host"
        expected_port = 54321
        expected = {
            "host": expected_host,
            "port": expected_port,
        }
        template_values.update({
            "mode": "manual",
            scheme + "_host": expected_host,
            scheme + "_port": expected_port,
        })
        fake_output = TEMPLATE_GSETTINGS_OUTPUT.format(**template_values)
        self.patch(gsettings.subprocess, "check_output",
                   lambda _: fake_output)
        ps = gsettings.get_proxy_settings()
        self.assertEqual(ps[scheme], expected)

    def test_gsettings_parser_http_anonymous(self):
        """Test a parser of gsettings."""
        self._assert_parser_anonymous('http')

    def test_gsettings_parser_https_anonymus(self):
        """Test a parser of gsettings."""
        self._assert_parser_anonymous('https')

    def test_gsettings_empty_ignore_hosts(self):
        """Missing values in the ignore hosts."""
        troublesome_value = "@as []"
        template_values = dict(BASE_GSETTINGS_VALUES)
        template_values["ignore_hosts"] = troublesome_value
        fake_output = TEMPLATE_GSETTINGS_OUTPUT.format(**template_values)
        self.patch(gsettings.subprocess, "check_output",
                   lambda _: fake_output)
        ps = gsettings.get_proxy_settings()
        self.assertEqual(ps, {})

    def test_gsettings_cannot_parse(self):
        """Some weird setting that cannot be parsed is logged with warning."""
        memento = MementoHandler()
        memento.setLevel(logging.DEBUG)
        gsettings.logger.addHandler(memento)
        self.addCleanup(gsettings.logger.removeHandler, memento)

        troublesome_value = "#bang"
        template_values = dict(BASE_GSETTINGS_VALUES)
        template_values["ignore_hosts"] = troublesome_value
        fake_output = TEMPLATE_GSETTINGS_OUTPUT.format(**template_values)
        self.patch(gsettings.subprocess, "check_output",
                   lambda _: fake_output)
        ps = gsettings.get_proxy_settings()
        self.assertTrue(memento.check_warning(gsettings.CANNOT_PARSE_WARNING %
                                              troublesome_value))
        self.assertEqual(ps, {})

    def test_gsettings_parser_http_authenticated(self):
        """Test a parser of gsettings."""
        template_values = dict(BASE_GSETTINGS_VALUES)
        expected_host = "expected_host"
        expected_port = 54321
        expected_user = "carlitos"
        expected_password = "very secret password"
        expected = {
            "host": expected_host,
            "port": expected_port,
            "username": expected_user,
            "password": expected_password,
        }
        template_values.update({
            "mode": "manual",
            "http_host": expected_host,
            "http_port": expected_port,
            "auth_user": expected_user,
            "auth_password": expected_password,
            "http_use_auth": "true",
        })
        fake_output = TEMPLATE_GSETTINGS_OUTPUT.format(**template_values)
        self.patch(gsettings.subprocess, "check_output",
                   lambda _: fake_output)
        ps = gsettings.get_proxy_settings()
        self.assertEqual(ps["http"], expected)

    def _assert_parser_authenticated_url(self, scheme):
        """Test a parser of gsettings with creds in the url."""
        template_values = dict(BASE_GSETTINGS_VALUES)
        expected_host = "expected_host"
        expected_port = 54321
        expected_user = "carlitos"
        expected_password = "very secret password"
        composed_url = '%s:%s@%s' % (expected_user, expected_password,
                                     expected_host)
        expected = {
            "host": expected_host,
            "port": expected_port,
            "username": expected_user,
            "password": expected_password,
        }
        template_values.update({
            "mode": "manual",
            scheme + "_host": composed_url,
            scheme + "_port": expected_port,
            "http_use_auth": "false",
        })
        fake_output = TEMPLATE_GSETTINGS_OUTPUT.format(**template_values)
        self.patch(gsettings.subprocess, "check_output",
                   lambda _: fake_output)
        ps = gsettings.get_proxy_settings()
        self.assertEqual(ps[scheme], expected)

    def test_gsettings_parser_http_authenticated_url(self):
        """Test a parser of gsettings with creds in the url."""
        self._assert_parser_authenticated_url('http')

    def test_gsettings_parser_https_authenticated_url(self):
        """Test a parser of gsettings with creds in the url."""
        self._assert_parser_authenticated_url('https')

    def test_gsettings_auth_over_url(self):
        """Test that the settings are more important that the url."""
        template_values = dict(BASE_GSETTINGS_VALUES)
        expected_host = "expected_host"
        expected_port = 54321
        expected_user = "carlitos"
        expected_password = "very secret password"
        composed_url = '%s:%s@%s' % ('user', 'random',
                                     expected_host)
        http_expected = {
            "host": expected_host,
            "port": expected_port,
            "username": expected_user,
            "password": expected_password,
        }
        template_values.update({
            "mode": "manual",
            "http_host": composed_url,
            "http_port": expected_port,
            "auth_user": expected_user,
            "auth_password": expected_password,
            "http_use_auth": "true",
        })
        fake_output = TEMPLATE_GSETTINGS_OUTPUT.format(**template_values)
        self.patch(gsettings.subprocess, "check_output",
                   lambda _: fake_output)
        ps = gsettings.get_proxy_settings()
        self.assertEqual(ps["http"], http_expected)

    def _assert_parser_empty_url(self, scheme):
        """Assert the parsing of an empty url."""
        template_values = dict(BASE_GSETTINGS_VALUES)
        template_values.update({
            "mode": "manual",
            scheme + "_host": '',
            scheme + "_port": 0,
            "http_use_auth": "false",
        })
        fake_output = TEMPLATE_GSETTINGS_OUTPUT.format(**template_values)
        self.patch(gsettings.subprocess, "check_output",
                   lambda _: fake_output)
        ps = gsettings.get_proxy_settings()
        self.assertNotIn(scheme, ps)

    def test_gsettings_parser_empty_http_url(self):
        """Test when there is no http proxy set."""
        self._assert_parser_empty_url('http')

    def test_gsettings_parser_empty_https_url(self):
        """Test when there is no https proxy set."""
        self._assert_parser_empty_url('https')


class ParseProxyHostTestCase(TestCase):
    """Test the parsing of the domain."""

    def test_onlyhost(self):
        """Parse a host with no username or password."""
        sample = "hostname"
        hostname, username, password = gsettings.parse_proxy_host(sample)
        self.assertEqual(username, None)
        self.assertEqual(password, None)
        self.assertEqual(hostname, "hostname")

    def test_user_and_host(self):
        """Parse host just with the username."""
        sample = "username@hostname"
        hostname, username, password = gsettings.parse_proxy_host(sample)
        self.assertEqual(username, "username")
        self.assertEqual(password, None)
        self.assertEqual(hostname, "hostname")

    def test_user_pass_and_host(self):
        """Test parsing a host with a username and password."""
        sample = "username:password@hostname"
        hostname, username, password = gsettings.parse_proxy_host(sample)
        self.assertEqual(username, "username")
        self.assertEqual(password, "password")
        self.assertEqual(hostname, "hostname")

    def test_username_with_at(self):
        """Test parsing the host with a username with @."""
        sample = "username@company.com:password@hostname"
        hostname, username, password = gsettings.parse_proxy_host(sample)
        self.assertEqual(username, "username@company.com")
        self.assertEqual(password, "password")
        self.assertEqual(hostname, "hostname")

    def test_username_with_at_nopass(self):
        """Test parsing the host without a password."""
        sample = "username@company.com@hostname"
        hostname, username, password = gsettings.parse_proxy_host(sample)
        self.assertEqual(username, "username@company.com")
        self.assertEqual(password, None)
        self.assertEqual(hostname, "hostname")

    def test_user_pass_with_colon_and_host(self):
        """Test parsing the host with a password that contains :."""
        sample = "username:pass:word@hostname"
        hostname, username, password = gsettings.parse_proxy_host(sample)
        self.assertEqual(username, "username")
        self.assertEqual(password, "pass:word")
        self.assertEqual(hostname, "hostname")



# -*- coding: utf-8 -*-
#
# Copyright 2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Tests for the platform specific functions (for Windows)."""

import thread

from twisted.internet import defer
from twisted.trial.unittest import TestCase

from ubuntu_kylin_sso.utils.qtwisted import qtDeferToThread


class FakeException(Exception):
    """A Fake exception."""


class QtDeferToThreadTestCase(TestCase):
    """A Qt-based implementation of deferToThread."""

    @defer.inlineCallbacks
    def test_executes_function_returns_result(self):
        """The passed function is executed and returns the result."""

        def fake_operation(first, second="world!"):
            """Testing positional and keyword arguments."""
            return first + " " + second

        result = yield qtDeferToThread(fake_operation, "hola", second="mundo!")
        self.assertEqual(result, "hola mundo!")

    @defer.inlineCallbacks
    def test_exceptions_become_failures(self):
        """Any exception while running the function becomes a failure."""

        def fake_raise():
            """Throw the fake exception."""
            raise FakeException

        yield self.assertFailure(qtDeferToThread(fake_raise), FakeException)

    @defer.inlineCallbacks
    def test_function_runs_in_new_thread(self):
        """The concerned function is run in a different thread."""

        def inner_function():
            """This should be run in a different thread."""
            return thread.get_ident()

        other_thread_id = yield qtDeferToThread(inner_function)
        this_thread_id = thread.get_ident()
        self.assertNotEqual(this_thread_id, other_thread_id)



# -*- coding: utf-8 -*-
#
# Copyright 2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Email Verification page UI."""

from functools import partial

from PyQt4 import QtGui, QtCore

from ubuntu_kylin_sso import NO_OP
from ubuntu_kylin_sso.logger import setup_gui_logging
from ubuntu_kylin_sso.qt import build_general_error_message
from ubuntu_kylin_sso.qt.sso_wizard_page import SSOWizardPage
from ubuntu_kylin_sso.qt.ui.email_verification_ui import Ui_EmailVerificationPage
from ubuntu_kylin_sso.utils import compat
from ubuntu_kylin_sso.utils.ui import (
    ERROR_EMAIL_TOKEN,
    NEXT,
    VERIFICATION_CODE,
    VERIFY_EMAIL_TITLE,
    VERIFY_EMAIL_CONTENT,
)


logger = setup_gui_logging('ubuntu_kylin_sso.email_verification_page')


class EmailVerificationPage(SSOWizardPage):
    """Widget used to input the email verification code."""

    ui_class = Ui_EmailVerificationPage
    registrationSuccess = QtCore.pyqtSignal(compat.text_type)

    def __init__(self, *args, **kwargs):
        self.email = ''
        self.password = ''
        super(EmailVerificationPage, self).__init__(*args, **kwargs)

    @property
    def _signals(self):
        """The signals to connect to the backend."""
        result = {
            'EmailValidated':
                self._filter_by_app_name(self.on_email_validated),
            'EmailValidationError':
                self._filter_by_app_name(self.on_email_validation_error),
        }
        return result

    @property
    def verification_code(self):
        """Return the content of the verification code edit."""
        return str(self.ui.verification_code_edit.text())

    @property
    def next_button(self):
        """Return the button that move to the next stage."""
        return self.ui.next_button

    def _connect_ui(self):
        """Set the connection of signals."""
        self.ui.verification_code_edit.textChanged.connect(
            self.validate_form)
        self.next_button.clicked.connect(self.validate_email)

    def validate_form(self):
        """Check the state of the form."""
        code = self.verification_code.strip()
        enabled = len(code) > 0
        self.next_button.setEnabled(enabled)

    def _set_translated_strings(self):
        """Set the different titles."""
        self.header.set_title(VERIFY_EMAIL_TITLE)
        self.header.set_subtitle(VERIFY_EMAIL_CONTENT % {
            "app_name": self.app_name,
            "email": self.email,
        })
        self.ui.label.setText(VERIFICATION_CODE)
        self.ui.next_button.setText(NEXT)

    def set_titles(self, email):
        """This class needs to have a public set_titles.

        Since the subtitle contains data that is only known after SetupAccount
        and _set_translated_strings is only called on initialization.
        """
        self._set_translated_strings()

    def validate_email(self):
        """Call the next action."""
        logger.debug('EmailVerificationController.validate_email for: %s',
            self.email)
        code = compat.text_type(self.ui.verification_code_edit.text())
        args = (self.app_name, self.email, self.password, code)
        self.hide_error()
        self.show_overlay()
        f = self.backend.validate_email

        logger.info('Calling validate_email with email %r, password <hidden>, '
                    'app_name %r and email_token %r.', self.email,
                    self.app_name, code)
        error_handler = partial(self._handle_error, f,
                                self.on_email_validation_error)
        f(*args, reply_handler=NO_OP, error_handler=error_handler)

    def on_email_validated(self, app_name, email):
        """Signal thrown after the email is validated."""
        logger.info('EmailVerificationController.on_email_validated for %s, '
                    'email: %s', app_name, email)
        self.hide_overlay()
        self.registrationSuccess.emit(self.email)

    def on_email_validation_error(self, app_name, error):
        """Signal thrown when there's a problem validating the email."""
        logger.error('Got error on email validation %s, error: %s',
            app_name, error)
        self.hide_overlay()
        msg = error.pop(ERROR_EMAIL_TOKEN, '')
        msg += build_general_error_message(error)
        self.show_error(msg)

    # pylint: disable=C0103

    def initializePage(self):
        """Called to prepare the page just before it is shown."""
        logger.debug('initializePage - About to show EmailVerificationPage')
        self.next_button.setDefault(True)
        self.next_button.setEnabled(False)
        self.wizard().setButtonLayout([QtGui.QWizard.Stretch])

    # pylint: enable=C0103



# -*- coding: utf-8 -*-
#
# Copyright 2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Loading animation over a widget."""

from __future__ import unicode_literals

from PyQt4 import QtGui, QtCore

from ubuntu_kylin_sso.qt.ui import loadingoverlay_ui
from ubuntu_kylin_sso.utils.ui import LOADING_OVERLAY

LOADING_STYLE = '<span style="font-size:x-large;">{0}</span>'


class LoadingOverlay(QtGui.QFrame):
    """The widget that shows a loading animation and disable the widget below.

    In order to have this working, the Widget which is going to use this
    overlay has to reimplement the resizeEvent as follows:

    def resizeEvent(self, event):
        self.overlay.resize(event.size())
        event.accept()

    """

    def __init__(self, parent=None):
        super(LoadingOverlay, self).__init__(parent=parent)
        self.ui = loadingoverlay_ui.Ui_Form()
        self.ui.setupUi(self)

        self.timer = None
        self.counter = 0
        self.orientation = False

        self.ui.label.setText(LOADING_STYLE.format(LOADING_OVERLAY))

    # Invalid name "paintEvent", "eventFilter", "showEvent", "timerEvent"
    # pylint: disable=C0103

    def paintEvent(self, event):
        """Paint over the widget to overlay its content."""
        painter = QtGui.QPainter()
        painter.begin(self)
        painter.setRenderHint(QtGui.QPainter.TextAntialiasing, True)
        painter.setRenderHint(QtGui.QPainter.Antialiasing, True)
        painter.fillRect(event.rect(), QtGui.QBrush(
            QtGui.QColor(255, 255, 255, 135)))
        painter.setPen(QtGui.QPen(QtCore.Qt.NoPen))
        painter.end()
        QtGui.QFrame.paintEvent(self, event)

    def eventFilter(self, obj, event):
        """Filter events from Frame content to draw the dot animation."""
        if getattr(self, 'ui', None) is not None and \
           obj == self.ui.frm_box and event.type() == QtCore.QEvent.Paint:
            painter = QtGui.QPainter()
            painter.begin(obj)
            painter.setRenderHint(QtGui.QPainter.Antialiasing, True)
            pos_x = self.ui.frm_box.width() / 3
            x_padding = pos_x / 5
            for i in range(5):
                if self.counter != i:
                    linear_gradient = QtGui.QLinearGradient(
                        pos_x + (x_padding * i),
                        self.ui.frm_box.height() / 2 + 10,
                        pos_x + (x_padding * i) + 15,
                        self.ui.frm_box.height() / 2 + 25)
                    linear_gradient.setColorAt(0, QtGui.QColor(205, 200, 198))
                    linear_gradient.setColorAt(1, QtGui.QColor(237, 237, 237))
                    painter.setBrush(QtGui.QBrush(linear_gradient))
                else:
                    linear_gradient = QtGui.QLinearGradient(
                        pos_x + (x_padding * i),
                        self.ui.frm_box.height() / 2 + 10,
                        pos_x + (x_padding * i) + 15,
                        self.ui.frm_box.height() / 2 + 25)
                    linear_gradient.setColorAt(0, QtGui.QColor(240, 67, 26))
                    linear_gradient.setColorAt(1, QtGui.QColor(255, 122, 53))
                    painter.setBrush(QtGui.QBrush(linear_gradient))
                painter.drawEllipse(
                    pos_x + (x_padding * i),
                    self.ui.frm_box.height() / 2 + 10,
                    15, 15)

            painter.end()
        return False

    def showEvent(self, event):
        """Start the dot animation."""
        self.ui.frm_box.installEventFilter(self)
        palette = QtGui.QPalette(self.palette())
        palette.setColor(palette.Background, QtCore.Qt.transparent)
        self.setPalette(palette)

        if not self.timer:
            self.timer = self.startTimer(200)

    def timerEvent(self, event):
        """Execute a loop to update the dot animation."""
        if self.counter in (0, 4):
            self.orientation = not self.orientation
        if self.orientation:
            self.counter += 1
        else:
            self.counter -= 1
        self.update()

    # pylint: enable=C0103



# -*- coding: utf-8 -*-
#
# Copyright 2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Utility to spawn another program from a mainloop."""

import sys

from twisted.internet import defer

from ubuntu_kylin_sso.logger import setup_logging


logger = setup_logging("ubuntu_kylin_sso.utils.runner")


class SpawnError(Exception):
    """Generic error when spawning processes."""


class FailedToStartError(SpawnError):
    """The process could not be spawned."""


def is_qt4_main_loop_installed():
    """Check if the Qt4 main loop is installed."""
    result = False

    if 'PyQt4' not in sys.modules:
        return result

    try:
        from PyQt4.QtCore import QCoreApplication
        result = QCoreApplication.instance() is not None
    except ImportError:
        pass

    return result


def spawn_program(args, use_reactor=False):
    """Spawn the program specified by 'args'.

    - 'args' should be a sequence of program arguments, the program to execute
    is normally the first item in 'args'.

    Return a deferred that will be fired when the execution of 'args' finishes,
    passing as the deferred result code the program return code.

    On error, the returned deferred will be errback'd.

    """
    logger.debug('spawn_program: requested to spawn %r.', repr(args))
    d = defer.Deferred()

    # HACK: The qt runner does have problem in mac os x and does not raise the
    # finish signal
    if use_reactor or sys.platform == 'darwin':
        from ubuntu_kylin_sso.utils.runner import tx
        source = tx
    elif is_qt4_main_loop_installed():
        from ubuntu_kylin_sso.utils.runner import qt
        source = qt
    else:
        from ubuntu_kylin_sso.utils.runner import glib
        source = glib

    logger.debug('Spawn source is %r.', source)

    def reply_handler(status):
        """Callback the returned deferred."""
        logger.debug('The program %r finished with status %r.', args, status)
        d.callback(status)

    def error_handler(msg, failed_to_start=False):
        """Errback the returned deferred."""
        if failed_to_start:
            msg = 'Process %r could not be started (%r).' % (args, msg)
            exc = FailedToStartError(msg)
        else:
            exc = SpawnError('Unspecified error (%r).' % msg)

        logger.error('The program %r could not be run: %r', args, exc)
        d.errback(exc)

    source.spawn_program(args, reply_handler, error_handler)
    return d



# -*- coding: utf-8 -*-
#
# Copyright 2011-2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""A proxy-enabled restful client."""

import json

try:
    # pylint: disable=E0611,F0401
    from urllib.parse import urlencode
    # pylint: enable=E0611,F0401
except ImportError:
    from urllib import urlencode

from twisted.internet import defer

from ubuntu_kylin_sso.logger import setup_logging
from ubuntu_kylin_sso.utils import webclient

logger = setup_logging("ubuntu_kylin_sso.utils.webclient.restful")

POST_HEADERS = {
    "content-type": "application/x-www-form-urlencoded",
}


class RestfulClient(object):
    """A proxy-enabled restful client."""

    def __init__(self, service_iri, username=None, password=None,
                 oauth_credentials=None):
        """Initialize this instance."""
        assert service_iri.endswith("/")
        self.service_iri = service_iri
        self.webclient = webclient.webclient_factory(username=username,
                                                     password=password,
                                                     oauth_sign_plain=True)
        self.oauth_credentials = oauth_credentials

    @defer.inlineCallbacks
    def restcall(self, method, **kwargs):
        """Make a restful call."""
        assert isinstance(method, unicode)
        params = {}
        for key, value in kwargs.items():
            if isinstance(value, basestring):
                assert isinstance(value, unicode)
            params[key] = json.dumps(value)
        namespace, operation = method.split(".")
        params["ws.op"] = operation
        encoded_args = urlencode(params)
        iri = self.service_iri + namespace
        creds = self.oauth_credentials
        logger.debug('Performing REST call to %r.', iri)
        result = yield self.webclient.request(iri, method="POST",
                                              oauth_credentials=creds,
                                              post_content=encoded_args,
                                              extra_headers=POST_HEADERS)
        try:
            response = json.loads(result.content)
        except:
            logger.exception('Can not load json from REST request response '
                             '(content is %r).', result.content)
            raise
        else:
            defer.returnValue(response)

    def shutdown(self):
        """Stop the webclient used by this class."""
        self.webclient.shutdown()



# -*- coding: utf-8 -*-
#
# Copyright 2011-2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Network status implementation on Windows."""


# pylint: disable=F0401
# distutils-extra: ignore-import=pythoncom,win32com.server.policy
# distutils-extra: ignore-import=win32com.client
import pythoncom
# pylint: enable=F0401
from ctypes import windll, byref
from ctypes.wintypes import DWORD
from threading import Thread

from twisted.internet import defer
# pylint: disable=F0401
from win32com.server.policy import DesignatedWrapPolicy
from win32com.client import Dispatch
# pylint: enable=F0401

from ubuntu_kylin_sso.networkstate import NetworkFailException
from ubuntu_kylin_sso.networkstate.networkstates import ONLINE, OFFLINE
from ubuntu_kylin_sso.logger import setup_logging

logger = setup_logging("ubuntu_kylin_sso.networkstate")

# naming errors are deliberated because we are following the COM naming to make
# it clear for later developers.
# pylint: disable=C0103

# from EventSys.h
PROGID_EventSystem = "EventSystem.EventSystem"
PROGID_EventSubscription = "EventSystem.EventSubscription"

# SENS (System Event Notification Service) values for the events,
# this events contain the uuid of the event, the name of the event to be used
# as well as the method name of the method in the ISesNetwork interface that
# will be executed for the event.
# For more info look at:
# http://msdn.microsoft.com/en-us/library/aa377384(v=vs.85).aspx

SUBSCRIPTION_NETALIVE = ('{cd1dcbd6-a14d-4823-a0d2-8473afde360f}',
                         'UbuntuOne Network Alive',
                         'ConnectionMade')

SUBSCRIPTION_NETALIVE_NOQOC = ('{a82f0e80-1305-400c-ba56-375ae04264a1}',
                               'UbuntuOne Net Alive No Info',
                               'ConnectionMadeNoQOCInfo')

SUBSCRIPTION_NETLOST = ('{45233130-b6c3-44fb-a6af-487c47cee611}',
                        'UbuntuOne Network Lost',
                        'ConnectionLost')

SUBSCRIPTION_REACH = ('{4c6b2afa-3235-4185-8558-57a7a922ac7b}',
                       'UbuntuOne Network Reach',
                       'ConnectionMade')

SUBSCRIPTION_REACH_NOQOC = ('{db62fa23-4c3e-47a3-aef2-b843016177cf}',
                            'UbuntuOne Network Reach No Info',
                            'ConnectionMadeNoQOCInfo')

SUBSCRIPTION_REACH_NOQOC2 = ('{d4d8097a-60c6-440d-a6da-918b619ae4b7}',
                             'UbuntuOne Network Reach No Info 2',
                             'ConnectionMadeNoQOCInfo')

SUBSCRIPTIONS = [SUBSCRIPTION_NETALIVE,
                 SUBSCRIPTION_NETALIVE_NOQOC,
                 SUBSCRIPTION_NETLOST,
                 SUBSCRIPTION_REACH,
                 SUBSCRIPTION_REACH_NOQOC,
                 SUBSCRIPTION_REACH_NOQOC2]

SENSGUID_EVENTCLASS_NETWORK = '{d5978620-5b9f-11d1-8dd2-00aa004abd5e}'
SENSGUID_PUBLISHER = "{5fee1bd6-5b9b-11d1-8dd2-00aa004abd5e}"

# uuid of the implemented com interface
IID_ISesNetwork = '{d597bab1-5b9f-11d1-8dd2-00aa004abd5e}'


class NetworkManager(DesignatedWrapPolicy):
    """Implement ISesNetwork to know about the network status."""

    _com_interfaces_ = [IID_ISesNetwork]
    _public_methods_ = ['ConnectionMade',
                        'ConnectionMadeNoQOCInfo',
                        'ConnectionLost']
    _reg_clsid_ = '{41B032DA-86B5-4907-A7F7-958E59333010}'
    _reg_progid_ = "UbuntuOne.NetworkManager"

    def __init__(self, connected_cb=None, connected_cb_info=None,
                 disconnected_cb=None):
        # pylint: disable=E1101,E1120,W0231
        self._wrap_(self)
        # pylint: enable=E1101,E1120
        self.connected_cb = connected_cb
        self.connected_cb_info = connected_cb_info
        self.disconnected_cb = disconnected_cb

    def ConnectionMade(self, *args):
        """Tell that the connection is up again."""
        logger.info('Connection was made.')
        if self.connected_cb_info:
            self.connected_cb_info()

    def ConnectionMadeNoQOCInfo(self, *args):
        """Tell that the connection is up again."""
        logger.info('Connection was made no info.')
        if self.connected_cb:
            self.connected_cb()

    def ConnectionLost(self, *args):
        """Tell the connection was lost."""
        logger.info('Connection was lost.')
        if self.disconnected_cb:
            self.disconnected_cb()

    def register(self):
        """Register to listen to network events."""
        # call the CoInitialize to allow the registration to run in another
        # thread
        # pylint: disable=E1101
        pythoncom.CoInitialize()
        # interface to be used by com
        manager_interface = pythoncom.WrapObject(self)
        event_system = Dispatch(PROGID_EventSystem)
        # register to listen to each of the events to make sure that
        # the code will work on all platforms.
        for current_event in SUBSCRIPTIONS:
            # create an event subscription and add it to the event
            # service
            event_subscription = Dispatch(PROGID_EventSubscription)
            event_subscription.EventClassId = SENSGUID_EVENTCLASS_NETWORK
            event_subscription.PublisherID = SENSGUID_PUBLISHER
            event_subscription.SubscriptionID = current_event[0]
            event_subscription.SubscriptionName = current_event[1]
            event_subscription.MethodName = current_event[2]
            event_subscription.SubscriberInterface = manager_interface
            event_subscription.PerUser = True
            # store the event
            try:
                event_system.Store(PROGID_EventSubscription,
                                   event_subscription)
            except pythoncom.com_error as e:
                logger.error(
                    'Error registering %s to event %s', e, current_event[1])

        pythoncom.PumpMessages()
        # pylint: enable=E1101


def is_machine_connected():
    """Return a deferred that when fired, returns if the machine is online."""
    # pylint: disable=W0703, W0702
    try:
        wininet = windll.wininet
        flags = DWORD()
        connected = wininet.InternetGetConnectedState(byref(flags), None)
        return defer.succeed(connected == 1)
    except Exception as e:
        logger.exception('is_machine_connected failed with:')
        return defer.fail(NetworkFailException(e))
    # pylint: enable=W0703, W0702


class NetworkManagerState(object):
    """Check for status changed in the network on Windows."""

    def __init__(self, result_cb, **kwargs):
        """Initialize this instance with a result and error callbacks."""
        self.result_cb = result_cb

    def connection_made(self):
        """Return the connection state over the call back."""
        self.result_cb(ONLINE)

    def connection_lost(self):
        """Return the connection was lost over the call back."""
        self.result_cb(OFFLINE)

    def find_online_state(self, listener=None, listener_thread=None):
        """Get the network state and return it thru the set callback."""
        # check the current status right now
        if is_machine_connected():
            self.result_cb(ONLINE)
        else:
            self.result_cb(OFFLINE)
        if listener is None:
            # start listening for network changes
            listener = NetworkManager(connected_cb=self.connection_made,
                                      disconnected_cb=self.connection_lost)
        if listener_thread is None:
            listener_thread = Thread(target=listener.register,
                                     name='Ubuntu SSO NetworkManager')
        listener_thread.daemon = True
        listener_thread.start()



# -*- coding: utf-8 -*-
#
# Copyright 2011-2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Tests for the proxy-aware webclient."""

TEMPLATE_GSETTINGS_OUTPUT = """\
org.gnome.system.proxy autoconfig-url '{autoconfig_url}'
org.gnome.system.proxy ignore-hosts {ignore_hosts:s}
org.gnome.system.proxy mode '{mode}'
org.gnome.system.proxy.ftp host '{ftp_host}'
org.gnome.system.proxy.ftp port {ftp_port}
org.gnome.system.proxy.http authentication-password '{auth_password}'
org.gnome.system.proxy.http authentication-user '{auth_user}'
org.gnome.system.proxy.http host '{http_host}'
org.gnome.system.proxy.http port {http_port}
org.gnome.system.proxy.http use-authentication {http_use_auth}
org.gnome.system.proxy.https host '{https_host}'
org.gnome.system.proxy.https port {https_port}
org.gnome.system.proxy.socks host '{socks_host}'
org.gnome.system.proxy.socks port {socks_port}
"""

BASE_GSETTINGS_VALUES = {
    "autoconfig_url": "",
    "ignore_hosts": ["localhost", "127.0.0.0/8"],
    "mode": "none",
    "ftp_host": "",
    "ftp_port": 0,
    "auth_password": "",
    "auth_user": "",
    "http_host": "",
    "http_port": 0,
    "http_use_auth": "false",
    "https_host": "",
    "https_port": 0,
    "socks_host": "",
    "socks_port": 0,
}



# -*- coding: utf-8 -*-
#
# Copyright 2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Test Ubunto SSO Wizard."""

from PyQt4 import QtGui
from twisted.internet import defer

from ubuntu_kylin_sso.tests import (
    APP_NAME,
    EMAIL,
    PASSWORD,
)
from ubuntu_kylin_sso.qt import PREFERED_UI_SIZE, ubuntu_sso_wizard
from ubuntu_kylin_sso.qt.tests import (
    BaseTestCase,
    FakeOverlay,
)
from ubuntu_kylin_sso.utils import compat


# is ok to access private method/attrs in tests
# pylint: disable=W0212


class UbuntuSSOClientGUITestCase(BaseTestCase):

    """Test the UbuntuSSOClientGUI class."""

    ui_class = ubuntu_sso_wizard.UbuntuSSOClientGUI
    ui_wizard_class = None

    @defer.inlineCallbacks
    def setUp(self):
        self.patch(ubuntu_sso_wizard.networkstate, 'is_machine_connected',
            lambda: True)
        yield super(UbuntuSSOClientGUITestCase, self).setUp()
        self.patch(ubuntu_sso_wizard.sys, 'exit', self._set_called)

    def test_closing_main_window(self):
        """When closing the main window, USER_CANCELLATION is returned."""
        self.ui.close()
        self.assertEqual(self._called,
            ((ubuntu_sso_wizard.USER_CANCELLATION,), {}))

    def test_close_main_window_reaching_success_page(self):
        """Closing the window in success page, USER_SUCCESS is returned."""
        self.ui.wizard._move_to_success_page()
        self.ui.close()
        self.assertEqual(self._called,
            ((ubuntu_sso_wizard.USER_SUCCESS,), {}))

    def test_closing_wizard_no_parent(self):
        """When closing the wizard, USER_CANCELLATION is returned."""
        self.ui.wizard.setParent(None)
        self.ui.wizard.close()
        self.assertEqual(self._called,
            ((ubuntu_sso_wizard.USER_CANCELLATION,), {}))

    def test_close_wizard_no_parent_reaching_success_page(self):
        """Closing the wizard in success page, USER_SUCCESS is returned."""
        self.ui.wizard.setParent(None)
        self.ui.wizard._move_to_success_page()
        self.ui.wizard.close()
        self.assertEqual(self._called,
            ((ubuntu_sso_wizard.USER_SUCCESS,), {}))

    def test_on_wizard_finish_done_called(self):
        """Test that the wizard actually close the window on finish."""
        self.ui.wizard._move_to_success_page()
        finish_button = self.ui.wizard.button(QtGui.QWizard.FinishButton)
        self.patch(self.ui.wizard, 'done', self._set_called)
        finish_button.clicked.emit(True)
        self.assertEqual(self._called, ((1,), {}))

    def test_on_wizard_finish_close_called_on_done(self):
        """Test that the wizard actually close the window on finish."""
        self.ui.wizard._move_to_success_page()
        finish_button = self.ui.wizard.button(QtGui.QWizard.FinishButton)
        self.patch(self.ui.wizard, 'closeEvent', self._set_called)
        finish_button.clicked.emit(True)
        self.assertEqual(self._called, ((None,), {}))

    def test_window_title(self):
        """Check the window title for the application."""
        title = compat.text_type(self.ui.windowTitle())
        self.assertEqual(title, ubuntu_sso_wizard.WINDOW_TITLE)


class UbuntuSSOWizardTestCase(BaseTestCase):

    """Test case for UbuntuSSOWizard class."""

    ui_class = ubuntu_sso_wizard.UbuntuSSOWizard
    ui_signals = ('recoverableError', 'loginSuccess', 'registrationSuccess')
    ui_wizard_class = None

    @defer.inlineCallbacks
    def setUp(self):
        self.patch(ubuntu_sso_wizard.networkstate, 'is_machine_connected',
            lambda: True)
        self.patch(ubuntu_sso_wizard, 'LoadingOverlay', FakeOverlay)
        yield super(UbuntuSSOWizardTestCase, self).setUp()

    def test_window_size(self):
        """check the window size."""
        size = self.ui.size()
        self.assertTrue(size.height() >= PREFERED_UI_SIZE['height'])
        self.assertTrue(size.width() >= PREFERED_UI_SIZE['width'])

    def test_move_to_success_page_state(self):
        """Test _move_to_success_page method."""
        self.patch(self.ui, 'setButtonLayout', self._set_called)
        self.ui._move_to_success_page()
        self.assertEqual(self.ui._next_id, -1)
        self.assertEqual(self.ui._next_id, -1)
        self.assertEqual(self._called,
            (([QtGui.QWizard.Stretch, QtGui.QWizard.FinishButton],), {}))
        self.assertTrue(self.ui.button(QtGui.QWizard.FinishButton).isEnabled())
        self.assertEqual(self.ui.exit_code, ubuntu_sso_wizard.USER_SUCCESS)

    def test_overlay_shows(self):
        """Test if the signals call the overlay.show properly."""
        # reset the counter
        self.ui.overlay.show_counter = 0

        for page in self.ui._pages:
            page.show_overlay()

        self.assertEqual(self.ui.overlay.show_counter, len(self.ui._pages))

    def test_overlay_hides(self):
        """Test if the signals call the overlay.show properly."""
        # reset the counter
        self.ui.overlay.show_counter = 0

        for page in self.ui._pages:
            page.hide_overlay()

        # We do +1 because a hide action is executed on the wizard init.
        self.assertEqual(self.ui.overlay.hide_counter,
            len(self.ui._pages) + 1)

    def test_go_back_to_current_user_sign_in(self):
        """Move the wizard back to the current user sign in page."""
        self.patch(self.ui, '_go_back_to_page', self._set_called)
        self.ui._next_id = self.ui.current_user_page_id
        self.ui.next()
        self.ui._move_to_reset_password_page()
        self.ui.reset_password.passwordChanged.emit('')
        expected = ((self.ui.current_user,), {})
        self.assertEqual(expected, self._called)

    def test_email_verification_page_params_from_current_user(self):
        """Tests that email_verification_page receives the proper params."""
        self.ui._next_id = self.ui.current_user_page_id
        self.ui.next()
        self.ui.current_user.ui.email_edit.setText(EMAIL)
        self.ui.current_user.ui.password_edit.setText(PASSWORD)
        self.ui.current_user.on_user_not_validated(APP_NAME, EMAIL)
        self.assertEqual(EMAIL, self.ui.email_verification.email)
        self.assertEqual(PASSWORD, self.ui.email_verification.password)

    def test_email_verification_page_params_from_setup(self):
        """Tests that email_verification_page receives the proper params."""
        self.ui._next_id = self.ui.setup_account_page_id
        self.ui.next()
        self.ui.setup_account.ui.email_edit.setText(EMAIL)
        self.ui.setup_account.ui.password_edit.setText(PASSWORD)
        self.ui.setup_account.on_user_registered(APP_NAME, {})
        self.assertEqual(EMAIL, self.ui.email_verification.email)
        self.assertEqual(PASSWORD, self.ui.email_verification.password)

    def test_check_network_connection(self):
        """Test that _connection_detected is executed with connection."""
        self.patch(self.ui, "_connection_detected", self._set_called)
        self.ui.check_network_connection()
        self.assertEqual(self._called, ((), {}))

    def test_check_network_connection_without_connection(self):
        """Test that _connection_detected is executed without connection."""
        self.patch(self.ui, "_connection_detected", self._set_called)
        self.patch(ubuntu_sso_wizard.networkstate, 'is_machine_connected',
            lambda: False)
        self.ui.check_network_connection()
        self.assertFalse(self._called)



# -*- coding: utf-8 -*-
#
# Copyright 2011-2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Perspective Broker IPC utils."""

from functools import wraps, partial
from collections import defaultdict

from twisted.internet import defer, endpoints
from twisted.spread.pb import (
    DeadReferenceError,
    NoSuchMethod,
    PBClientFactory,
    PBServerFactory,
    Referenceable,
    Root,
)

from ubuntu_kylin_sso.logger import setup_logging
from ubuntu_kylin_sso.utils.tcpactivation import (
    ActivationClient,
    ActivationConfig,
    ActivationInstance,
)


logger = setup_logging("ubuntu_kylin_sso.utils.ipc")
LOCALHOST = '127.0.0.1'

# pylint: disable=E1103


@defer.inlineCallbacks
def server_listen(server_factory, service_name, activation_cmdline,
                  description, reactor=None):
    """Connect the IPC server factory."""
    config = ActivationConfig(service_name, activation_cmdline, description)
    ai = ActivationInstance(config)
    description = yield ai.get_server_description()

    if reactor is None:
        from twisted.internet import reactor
    server = endpoints.serverFromString(reactor, description)
    connector = yield server.listen(server_factory)
    defer.returnValue(connector)


@defer.inlineCallbacks
def client_connect(client_factory, service_name, activation_cmdline,
                   description, reactor=None):
    """Connect the IPC client factory."""
    config = ActivationConfig(service_name, activation_cmdline, description)
    ac = ActivationClient(config)
    description = yield ac.get_active_client_description()

    if reactor is None:
        from twisted.internet import reactor

    client = endpoints.clientFromString(reactor, description)
    port = yield client.connect(client_factory)
    defer.returnValue(port)


# pylint: enable=E1103

# ============================== service helpers ==============================

def signal(f):
    """Decorator to emit a signal."""

    @wraps(f)
    def inner(self, *args, **kwargs):
        """Grab the signal name from the internal mapping and emit."""
        f(self, *args, **kwargs)
        signal_name = f.__name__
        self.emit_signal(signal_name, *args, **kwargs)

    return inner


class SignalBroadcaster(object):
    """Object that allows to emit signals to clients over the IPC."""

    MSG_NO_SIGNAL_HANDLER = "No signal handler for %r in %r"
    MSG_COULD_NOT_EMIT_SIGNAL = "Could not emit signal %r to %r due to %r"

    def __init__(self):
        """Create a new instance."""
        super(SignalBroadcaster, self).__init__()
        self.clients_per_signal = defaultdict(set)

    def _ignore_no_such_method(self, failure, signal_name, current_client):
        """NoSuchMethod is not an error, ignore it."""
        failure.trap(NoSuchMethod)
        logger.debug(self.MSG_NO_SIGNAL_HANDLER, signal_name, current_client)

    def _other_failure(self, failure, signal_name, current_client):
        """Log the issue when emitting a signal."""
        logger.warning(self.MSG_COULD_NOT_EMIT_SIGNAL, signal_name,
                       current_client, failure.value)
        logger.warning('Traceback is:\n%s', failure.printDetailedTraceback())

    def remote_register_to_signals(self, client, signals):
        """Allow a client to register to some signals."""
        for signal_name in signals:
            self.clients_per_signal[signal_name].add(client)

    def remote_unregister_to_signals(self, client):
        """Allow a client to unregister from the signal."""
        for connected_clients in self.clients_per_signal.values():
            if client in connected_clients:
                connected_clients.remove(client)

    def emit_signal(self, signal_name, *args, **kwargs):
        """Emit the given signal to the clients."""
        logger.debug("emitting %r to all (%i) connected clients.",
                     signal_name, len(self.clients_per_signal[signal_name]))
        result = []
        dead_clients = set()
        for current_client in self.clients_per_signal[signal_name]:
            try:
                d = current_client.callRemote(signal_name, *args, **kwargs)
                d.addErrback(self._ignore_no_such_method, signal_name,
                                                          current_client)
                d.addErrback(self._other_failure, signal_name, current_client)
                result.append(d)
            except DeadReferenceError:
                dead_clients.add(current_client)
        for client in dead_clients:
            logger.warn('emit_signal: Unregistering dead client %s.', client)
            self.remote_unregister_to_signals(client)


class RemoteMeta(type):
    """Append remote_ to the remote methods.

    Remote has to be appended to the remote method to work over pb but this
    names cannot be used since the other platforms do not expect the remote
    prefix. This metaclass creates those prefixes so that the methods can be
    correctly called.
    """

    def __new__(mcs, name, bases, attrs):
        remote_calls = attrs.get('remote_calls', [])
        for current in remote_calls:
            attrs['remote_' + current] = attrs[current]
        return super(RemoteMeta, mcs).__new__(mcs, name, bases, attrs)


def meta_base(meta):
    """A function to return a metaclass which is called directly.
    This is safe for both Python 2 and 3."""
    return meta("MetaBase", (object,), {})


class RemoteService(meta_base(RemoteMeta),
                    Referenceable, SignalBroadcaster):
    """A remote service that provides the methods listed in remote_calls."""

    remote_calls = []


class BaseService(object, Root):
    """Base PB service.

    Inherit from this class and define name, description and cmdline.

    If 'start' is called, 'shutdown' should be called when done.

    """

    # a mapping of (service name, service class (an instance of RemoteService))
    services = {}
    name = None
    description = None
    cmdline = None

    def __init__(self, *a, **kw):
        super(BaseService, self).__init__()
        self.factory = None
        self.listener = None
        for name, service_class in self.services.items():
            service = service_class(*a, **kw)
            setattr(self, name, service)
            setattr(self, 'remote_get_%s' % name,
                    partial(self._get_service, name))

    def _get_service(self, name):
        """Return the instance of the service named 'name'."""
        return getattr(self, name)

    @defer.inlineCallbacks
    def start(self):
        """Start listening in the proper description."""
        self.factory = PBServerFactory(self)
        self.listener = yield server_listen(self.factory,
                                            self.name,
                                            self.cmdline,
                                            self.description)

    def shutdown(self):
        """Stop listening."""
        self.listener.stopListening()


# ============================== client helpers ==============================


class RemoteClient(object, Referenceable):
    """Represent a client for remote calls."""

    call_remote_functions = []  # methods that can be called on the remote obj
    signal_handlers = []  # signals that are of interest of this client

    def __init__(self, base_client, remote_object):
        """Create instance."""
        super(RemoteClient, self).__init__()
        self._mapping = defaultdict(list)
        self.base_client = base_client
        self.remote = remote_object

        # for each function name in self.call_remote_functions,
        # set as instance attribute a function that will execute
        # self.remote.callRemote(function name, *a, **kw)
        for name in self.call_remote_functions:
            setattr(self, name, partial(self.call_method, name))

        # for each signal name in self.signal_handlers,
        # set as instance attribute a function that will execute
        # all the callbacks registered with connect_to_signal, plus
        # the method self.<signal_name>_cb(*a, **kw) if defined
        for name in self.signal_handlers:
            setattr(self, 'remote_' + name, self._emit_signal(name))

    def _emit_signal(self, fname):
        """Return a function that will execute the 'fname' method."""

        def callback_wrapper(*args, **kwargs):
            """Return the result of the callback if present."""
            for callback in self._mapping[fname]:
                logger.info('Emitting remote signal for %s with callback %r.',
                            fname, callback)
                callback(*args, **kwargs)

            # execute methods ending with _cb as signal handlers,
            # only for old compatibility
            old_name = ['on']
            for i in fname:
                if i.isupper():
                    old_name.append('_' + i.lower())
                else:
                    old_name.append(i)
            old_name.append('_cb')

            callback = getattr(self, ''.join(old_name), None)
            if callback is not None:
                logger.info('Emitting remote signal for %s with callback %r.',
                            fname, callback)
                callback(*args, **kwargs)

        return callback_wrapper

    @defer.inlineCallbacks
    def call_method(self, method_name, *args, **kwargs):
        """Call asynchronously 'method_name(*args)'.

        Return a deferred that will be fired when the call finishes.
        For now, **kwargs are ignored.

        """
        logger.debug('Performing %r as a remote call (%r, %r).',
                     method_name, args, kwargs)
        try:
            result = yield self.remote.callRemote(method_name, *args)
        except DeadReferenceError:
            yield self.base_client.reconnect()
            result = yield self.call_method(method_name, *args, **kwargs)
        defer.returnValue(result)

    @defer.inlineCallbacks
    def register_to_signals(self):
        """Register to the signals."""
        try:
            result = yield self.remote.callRemote('register_to_signals', self,
                    self.signal_handlers)
        except DeadReferenceError:
            yield self.base_client.reconnect()
            result = yield self.register_to_signals()
        defer.returnValue(result)

    def unregister_to_signals(self):
        """Register to the signals."""
        return self.remote.callRemote('unregister_to_signals', self)

    def connect_to_signal(self, signal_name, callback):
        """Register 'callback' to be called when 'signal_name' is emitted."""
        self._mapping[signal_name].append(callback)
        return callback

    def disconnect_from_signal(self, signal_name, match):
        """Unregister 'match' from 'signal_name'.

        'match' is the object returned by 'connect_to_signal'.

        """
        if match in self._mapping[signal_name]:
            self._mapping[signal_name].remove(match)


class BaseClient(object):
    """Client that will connect to the service listening on the description.

    Inherit from this class and define service_name, service_description and
    service_cmdline so they return the proper values.

    The service_cmdline must be redefined so it returns the command line to
    execute to run the service, if it's not running.

    If 'connect' is called, 'disconnect' should be called when done.

    """

    # a mapping of (client name, client class (an instance of RemoteClient))
    clients = {}
    service_name = None
    service_description = None
    service_cmdline = None

    def __init__(self):
        self.factory = None
        self.client = None
        for client in self.clients:
            setattr(self, client, None)

    @defer.inlineCallbacks
    def _request_remote_objects(self, root):
        """Request all the diff remote objects used for the communication."""
        logger.debug('Requesting remote objects (%r) for %s',
                     self.clients.keys(), self.__class__.__name__)
        for name, client_class in self.clients.items():
            remote = yield root.callRemote('get_%s' % name)
            setattr(self, name, client_class(self, remote))

    @defer.inlineCallbacks
    def connect(self):
        """Connect to the remote service."""
        self.factory = PBClientFactory()
        self.client = yield client_connect(self.factory,
                                           self.service_name,
                                           self.service_cmdline,
                                           self.service_description)
        root = yield self.factory.getRootObject()
        yield self._request_remote_objects(root)
        yield self.register_to_signals()

    @defer.inlineCallbacks
    def reconnect(self):
        """Reconnect with the server."""
        self.factory = PBClientFactory()
        self.client = yield client_connect(self.factory,
                                           self.service_name,
                                           self.service_cmdline,
                                           self.service_description)
        root = yield self.factory.getRootObject()
        # loop over the already present remote clients and reset their remotes
        for name in self.clients:
            remote = yield root.callRemote('get_%s' % name)
            remote_client = getattr(self, name)
            remote_client.remote = remote
        yield self.register_to_signals()

    @defer.inlineCallbacks
    def register_to_signals(self):
        """Register all the clients to their signals."""
        for name in self.clients:
            client = getattr(self, name)
            yield client.register_to_signals()

    @defer.inlineCallbacks
    def unregister_to_signals(self):
        """Unregister from the all the client's signals."""
        for name in self.clients:
            client = getattr(self, name)
            yield client.unregister_to_signals()

    @defer.inlineCallbacks
    def disconnect(self):
        """Disconnect from the process."""
        yield self.unregister_to_signals()
        if self.client:
            self.client.disconnect()



# -*- coding: utf-8 -*-
#
# Copyright 2011-2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Test the pykeyring implementation."""

from json import dumps

from twisted.internet import defer
from twisted.trial.unittest import TestCase

from ubuntu_kylin_sso.keyring import pykeyring


class FakePyKeyring(object):

    """Fake Keyring"""

    def __init__(self):
        """Initialize object."""
        self.data = {}

    def set_password(self, *args):
        """Fake set_password for keyring."""
        self.data['set_password'] = args
        self.data['app_name'] = args[0][0]
        return self

    def get_password(self, app_name, username):
        """Fake get_password."""
        self.data['get_password'] = True
        password = dict(password='password', app_name=app_name)
        return dumps(password)

    def delete_password(self, *args):
        """Fake delete_password."""
        self.data['delete_password'] = True


class FakeDefer(object):

    """Fake Deferred."""

    def fake_defer_to_thread(self, func, *args):
        """Fake deferToThread."""
        return func(args)


class TestPyKeyring(TestCase):
    """Test the pykeyring implementation."""

    @defer.inlineCallbacks
    def setUp(self):
        """Setup  tests."""
        yield super(TestPyKeyring, self).setUp()
        self.app_name = 'app_name'
        self.fake_keyring = FakePyKeyring()
        self.keyring = pykeyring.Keyring(self.fake_keyring)
        self.fake_defer = FakeDefer()
        self.patch(pykeyring, "deferToThread",
            self.fake_defer.fake_defer_to_thread)

    # pylint: disable=E1103
    def test_set_credentials(self):
        """Test setting the credentials."""
        self.keyring = pykeyring.Keyring(self.fake_keyring)
        d = self.keyring.set_credentials(self.app_name, 'password')
        self.assertEqual(d.data['set_password'],
            (('app_name', 'ubuntu_kylin_sso', '"password"'), ))

    def test_get_credentials(self):
        """Test getting the credentials."""
        d = self.keyring.get_credentials(self.app_name)
        self.assertTrue('password' in d)
        self.assertEqual(d['password'], 'password')
        self.assertTrue('app_name' in d)
        self.assertEqual(d['app_name'], ['app_name'])
        self.assertTrue(self.fake_keyring.data.get('get_password', False))

    def test_get_credentials_not_present(self):
        """Test getting creds that are not present."""
        d = self.keyring.get_credentials('app_name_not_there')
        self.assertTrue('password' in d)
        self.assertEqual(d['password'], 'password')
        self.assertTrue('app_name' in d)
        self.assertNotEqual(d['app_name'], [self.app_name])
        self.assertTrue(self.fake_keyring.data.get('get_password', False))

    def test_delete_credentials(self):
        """Test deleting the credentials."""
        result = self.keyring.delete_credentials(self.app_name)
        self.assertTrue(result is None)
        self.assertTrue(self.fake_keyring.data.get('delete_password', False))
    # pylint: enable=E1103



# -*- coding: utf-8 -*-
# Author: Diego Sarmentero <diego.sarmentero@canonical.com>
#
# Copyright 2011-2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Test the EnhancedLineEdit."""

import collections

from PyQt4 import QtGui, QtCore
from twisted.trial.unittest import TestCase

from ubuntu_kylin_sso.qt.sso_wizard_page import EnhancedLineEdit


class EnhancedLineEditTestCase(TestCase):
    """Tests EnhancedLineEdit for valid and invalid states."""

    def get_pixmap_data(self, pixmap):
        """Get the raw data of a QPixmap."""
        byte_array = QtCore.QByteArray()
        array_buffer = QtCore.QBuffer(byte_array)
        pixmap.save(array_buffer, "PNG")
        return byte_array

    # pylint: disable=C0103
    def assertEqualPixmaps(self, pixmap1, pixmap2):
        """Compare two Qt pixmaps."""
        d1 = self.get_pixmap_data(pixmap1)
        d2 = self.get_pixmap_data(pixmap2)
        self.assertEqual(d1, d2)
    # pylint: enable=C0103

    def test_enhanced_line_edit_init(self):
        """Tests the initial state of the enchanced line edit."""
        line_edit = QtGui.QLineEdit()
        line_edit.show()
        enhanced = EnhancedLineEdit(line_edit)
        self.assertIsInstance(enhanced.valid_cb, collections.Callable)
        self.assertTrue(enhanced.clear_label.property("lineEditWarning"))
        self.assertFalse(enhanced.clear_label.isVisible())
        self.assertEqual(enhanced.clear_label.cursor().shape(),
            QtCore.Qt.ArrowCursor)
        self.assertFalse(enhanced.clear_label.isVisible())
        self.assertEqual(enhanced.clear_label.pixmap(), None)

    def test_enhanced_line_edit_init_with_icon(self):
        """Tests the initial state of the enchanced line edit."""
        line_edit = QtGui.QLineEdit()
        line_edit.show()
        enhanced = EnhancedLineEdit(line_edit, warning_sign=True)
        self.assertIsInstance(enhanced.valid_cb, collections.Callable)
        edit_property = enhanced.clear_label.property("lineEditWarning")
        self.assertTrue(edit_property)
        self.assertFalse(enhanced.clear_label.isVisible())
        self.assertEqual(enhanced.clear_label.cursor().shape(),
            QtCore.Qt.ArrowCursor)
        self.assertFalse(enhanced.clear_label.isVisible())
        icon = QtGui.QApplication.style().standardIcon(
            QtGui.QStyle.SP_MessageBoxWarning)
        self.assertEqualPixmaps(enhanced.clear_label.pixmap(),
            icon.pixmap(16, 16))

    def test_enhanced_line_edit_function_show_true(self):
        """Tests the EnchancedLineEdit show method with validation = True."""
        line_edit = QtGui.QLineEdit()
        line_edit.show()
        func = lambda x: False
        enhanced = EnhancedLineEdit(line_edit, valid_cb=func,
                                    warning_sign=True)
        line_edit.setText("testing")
        self.assertTrue(enhanced.clear_label.isVisible())

    def test_enhanced_line_edit_function_show_false(self):
        """Tests the EnchancedLineEdit show method with validation = False."""
        line_edit = QtGui.QLineEdit()
        line_edit.show()
        func = lambda x: True
        enhanced = EnhancedLineEdit(line_edit, valid_cb=func)
        line_edit.setText("testing")
        self.assertFalse(enhanced.clear_label.isVisible())



# -*- coding: utf-8 -*-
#
# Copyright 2014 National University of Defense Technology(NUDT) & Kylin Ltd
# Copyright 2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Generic SSO client using Twisted Perspective Broker.

This module should never import from the multiplatform one (main/__init__.py),
but the other way around. Likewise, this module should *not* have any logic
regarding error processing or decision making about when to send a given
signal.

Also, most of the logging is being made in the main module to avoid
duplication between the different platform implementations.

"""

from twisted.internet.task import LoopingCall
from twisted.python.failure import Failure

from ubuntu_kylin_sso.logger import setup_logging
from ubuntu_kylin_sso.utils.ipc import (
    BaseService,
    RemoteClient,
    RemoteService,
    signal,
)

logger = setup_logging("ubuntu_kylin_sso.main.perspective_broker")
SSO_SERVICE_NAME = "ubuntu-kylin-sso-client"

# Invalid name for signals that are CamelCase
# pylint: disable=C0103


class SSOLoginProxy(RemoteService):
    """Login thru the Single Sign On service."""

    remote_calls = [
        'generate_captcha',
        'register_user',
        'login',
        'login_and_ping',
        'validate_email',
        'validate_email_and_ping',
        'request_password_reset_token',
        'set_new_password',
    ]

    def __init__(self, root, *args, **kwargs):
        super(SSOLoginProxy, self).__init__(*args, **kwargs)
        self.root = root

    # generate_capcha signals
    @signal
    def CaptchaGenerated(self, app_name, result):
        """Signal thrown after the captcha is generated."""

    @signal
    def CaptchaGenerationError(self, app_name, error):
        """Signal thrown when there's a problem generating the captcha."""

    def generate_captcha(self, app_name, filename):
        """Call the matching method in the processor."""
        self.root.sso_login.generate_captcha(app_name, filename)

    # register_user signals
    @signal
    def UserRegistered(self, app_name, result):
        """Signal thrown when the user is registered."""

    @signal
    def UserRegistrationError(self, app_name, error):
        """Signal thrown when there's a problem registering the user."""

    def register_user(self, app_name, email, password, name,
                      captcha_id, captcha_solution):
        """Call the matching method in the processor."""
        self.root.sso_login.register_user(app_name, email, password, name,
            captcha_id, captcha_solution)

    # login signals
    @signal
    def LoggedIn(self, app_name, result):
        """Signal thrown when the user is logged in."""

    @signal
    def LoginError(self, app_name, error):
        """Signal thrown when there is a problem in the login."""

    @signal
    def UserNotValidated(self, app_name, result):
        """Signal thrown when the user is not validated."""

    def login(self, app_name, email, password, ping_url=None):
        """Call the matching method in the processor."""
        self.root.sso_login.login(app_name, email, password, ping_url)

    login_and_ping = login

    # validate_email signals
    @signal
    def EmailValidated(self, app_name, result):
        """Signal thrown after the email is validated."""

    @signal
    def EmailValidationError(self, app_name, error):
        """Signal thrown when there's a problem validating the email."""

    def validate_email(self, app_name, email, password, email_token,
                       ping_url=None):
        """Call the matching method in the processor."""
        self.root.sso_login.validate_email(app_name,
            email, password, email_token, ping_url)

    validate_email_and_ping = validate_email

    # request_password_reset_token signals
    @signal
    def PasswordResetTokenSent(self, app_name, result):
        """Signal thrown when the token is succesfully sent."""

    @signal
    def PasswordResetError(self, app_name, error):
        """Signal thrown when there's a problem sending the token."""

    def request_password_reset_token(self, app_name, email):
        """Call the matching method in the processor."""
        self.root.sso_login.request_password_reset_token(app_name, email)

    # set_new_password signals
    @signal
    def PasswordChanged(self, app_name, result):
        """Signal thrown when the token is succesfully sent."""

    @signal
    def PasswordChangeError(self, app_name, error):
        """Signal thrown when there's a problem sending the token."""

    def set_new_password(self, app_name, email, token, new_password):
        """Call the matching method in the processor."""
        self.root.sso_login.set_new_password(app_name,
            email, token, new_password)


class CredentialsManagementProxy(RemoteService):
    """Object that manages credentials.

    Every exposed method in this class requires one mandatory argument:

        - 'app_name': the name of the application. Will be displayed in the
        GUI header, plus it will be used to find/build/clear tokens.

    And accepts another parameter named 'args', which is a dictionary that
    can contain the following:

        - 'help_text': an explanatory text for the end-users, will be
        shown below the header. This is an optional free text field.

        - 'ping_url': the url to open after successful token retrieval. If
        defined, the email will be attached to the url and will be pinged
        with a OAuth-signed request.

        - 'tc_url': the link to the Terms and Conditions page. If defined,
        the checkbox to agree to the terms will link to it.

        - 'window_id': the id of the window which will be set as a parent
        of the GUI. If not defined, no parent will be set.

    """

    remote_calls = [
        'find_credentials',
        'clear_credentials',
        'store_credentials',
        'register',
        'login',
        'login_email_password',
    ]

    def __init__(self, root, *args, **kwargs):
        super(CredentialsManagementProxy, self).__init__(*args, **kwargs)
        self.root = root

    @signal
    def AuthorizationDenied(self, app_name):
        """Signal thrown when the user denies the authorization."""

    @signal
    def CredentialsFound(self, app_name, credentials):
        """Signal thrown when the credentials are found."""

    @signal
    def CredentialsNotFound(self, app_name):
        """Signal thrown when the credentials are not found."""

    @signal
    def CredentialsCleared(self, app_name):
        """Signal thrown when the credentials were cleared."""

    @signal
    def CredentialsStored(self, app_name):
        """Signal thrown when the credentials were cleared."""

    @signal
    def CredentialsError(self, app_name, error_dict):
        """Signal thrown when there is a problem getting the credentials."""

    def find_credentials(self, app_name, args):
        """Look for the credentials for an application.

        - 'app_name': the name of the application which credentials are
        going to be removed.

        - 'args' is a dictionary, currently not used.

        """
        self.root.cred_manager.find_credentials(app_name, args)

    def clear_credentials(self, app_name, args):
        """Clear the credentials for an application.

        - 'app_name': the name of the application which credentials are
        going to be removed.

        - 'args' is a dictionary, currently not used.

        """
        self.root.cred_manager.clear_credentials(app_name, args)

    def store_credentials(self, app_name, args):
        """Store the token for an application.

        - 'app_name': the name of the application which credentials are
        going to be stored.

        - 'args' is the dictionary holding the credentials. Needs to provide
        the following mandatory keys: 'token', 'token_key', 'consumer_key',
        'consumer_secret'.

        """
        self.root.cred_manager.store_credentials(app_name, args)

    def register(self, app_name, args):
        """Get credentials if found else prompt GUI to register."""
        self.root.cred_manager.register(app_name, args)

    def login(self, app_name, args):
        """Get credentials if found else prompt GUI to login."""
        self.root.cred_manager.login(app_name, args)

    def login_email_password(self, app_name, args):
        """Get credentials if found, else login using email and password.

        - 'args' should contain at least the follwing keys: 'email' and
        'password'. Those will be used to issue a new SSO token, which will be
        returned trough the CredentialsFound signal.

        """
        self.root.cred_manager.login_email_password(app_name, args)


class UbuntuSSOProxyBase(BaseService):
    """Object that exposes the diff referenceable objects."""

    services = {
        'sso_login': SSOLoginProxy,
        'cred_manager': CredentialsManagementProxy,
    }

    name = SSO_SERVICE_NAME


# ============================== client classes ==============================


class SSOLoginClient(RemoteClient):
    """Client that can perform calls to the remote SSOLogin object."""

    call_remote_functions = SSOLoginProxy.remote_calls
    signal_handlers = [
        'CaptchaGenerated',
        'CaptchaGenerationError',
        'UserRegistered',
        'UserRegistrationError',
        'LoggedIn',
        'LoginError',
        'UserNotValidated',
        'EmailValidated',
        'EmailValidationError',
        'PasswordResetTokenSent',
        'PasswordResetError',
        'PasswordChanged',
        'PasswordChangeError',
    ]


class CredentialsManagementClient(RemoteClient):
    """Client that can perform calls to the remote CredManagement object."""

    call_remote_functions = CredentialsManagementProxy.remote_calls
    signal_handlers = [
        'AuthorizationDenied',
        'CredentialsFound',
        'CredentialsNotFound',
        'CredentialsCleared',
        'CredentialsStored',
        'CredentialsError',
    ]


def add_timeout(interval, callback, *args, **kwargs):
    """Add a timeout callback as a task."""
    time_out_task = LoopingCall(callback, *args, **kwargs)
    time_out_task.start(interval / 1000, now=False)


timeout_func = add_timeout
start_setup = lambda *a, **kw: None

# the reactor does have run and stop methods
# pylint: disable=E1101


def shutdown_func():
    """Stop the reactor."""
    from twisted.internet import reactor
    reactor.stop()


def finish_setup(result, loop):
    """Stop the reactor if a failure ocurred."""
    if isinstance(result, Failure):
        shutdown_func()


def main():
    """Run the specific mainloop."""
    from twisted.internet import reactor
    reactor.run()

# pylint: enable=E1101



# -*- coding: utf-8 -*-
#
# Copyright 2014 National University of Defense Technology(NUDT) & Kylin Ltd
# Copyright 2009-2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# In addition, as a special exception, the copyright holders give
# permission to link the code of portions of this program with the
# OpenSSL library under certain conditions as described in each
# individual source file, and distribute linked combinations
# including the two.
# You must obey the GNU General Public License in all respects
# for all of the code used other than OpenSSL.  If you modify
# file(s) with this exception, you may extend this exception to your
# version of the file(s), but you are not obligated to do so.  If you
# do not wish to do so, delete this exception statement from your
# version.  If you delete this exception statement from all source
# files in the program, then also delete it here.
"""Main module implementation specific for linux.

This module should never import from the multiplatform one (main/__init__.py),
but the other way around. Likewise, this module should *not* have any logic
regarding error processing or decision making about when to send a given
signal.

Also, most of the logging is being made in the main module to avoid
duplication between the different platform implementations.

"""

import signal

import dbus
import dbus.service

from twisted.internet import defer

from ubuntu_kylin_sso import (
    DBUS_ACCOUNT_PATH,
    DBUS_BUS_NAME,
    DBUS_CREDENTIALS_IFACE,
    DBUS_CREDENTIALS_PATH,
    DBUS_IFACE_USER_NAME,
    NO_OP,
)
from ubuntu_kylin_sso.logger import setup_logging


# Disable the invalid name warning, as we have a lot of DBus style names
# pylint: disable=C0103


logger = setup_logging("ubuntu_kylin_sso.main.linux")


class SSOLoginProxy(dbus.service.Object):
    """Login thru the Single Sign On service."""

    # Use of super on an old style class
    # pylint: disable=E1002

    def __init__(self, root, *args, **kwargs):
        """Initiate the Login object."""
        # pylint: disable=E1002
        super(SSOLoginProxy, self).__init__(*args, **kwargs)
        self.root = root

    # Operator not preceded by a space (fails with dbus decorators)
    # pylint: disable=C0322

    # generate_capcha signals
    @dbus.service.signal(DBUS_IFACE_USER_NAME, signature="ss")
    def CaptchaGenerated(self, app_name, result):
        """Signal thrown after the captcha is generated."""

    @dbus.service.signal(DBUS_IFACE_USER_NAME, signature="sa{ss}")
    def CaptchaGenerationError(self, app_name, error):
        """Signal thrown when there's a problem generating the captcha."""

    @dbus.service.method(dbus_interface=DBUS_IFACE_USER_NAME,
                         in_signature='ss')
    def generate_captcha(self, app_name, filename):
        """Call the matching method in the processor."""
        self.root.sso_login.generate_captcha(app_name, filename)

    # register_user signals
    @dbus.service.signal(DBUS_IFACE_USER_NAME, signature="ss")
    def UserRegistered(self, app_name, result):
        """Signal thrown when the user is registered."""

    @dbus.service.signal(DBUS_IFACE_USER_NAME, signature="sa{ss}")
    def UserRegistrationError(self, app_name, error):
        """Signal thrown when there's a problem registering the user."""

    @dbus.service.method(dbus_interface=DBUS_IFACE_USER_NAME,
                         in_signature='ssssss')
    def register_user(self, app_name, email, password, name,
                      captcha_id, captcha_solution):
        """Call the matching method in the processor."""
        self.root.sso_login.register_user(app_name, email, password, name,
            captcha_id, captcha_solution)

    # login signals
    @dbus.service.signal(DBUS_IFACE_USER_NAME, signature="ss")
    def LoggedIn(self, app_name, result):
        """Signal thrown when the user is logged in."""

    @dbus.service.signal(DBUS_IFACE_USER_NAME, signature="sa{ss}")
    def LoginError(self, app_name, error):
        """Signal thrown when there is a problem in the login."""

    @dbus.service.signal(DBUS_IFACE_USER_NAME, signature="ss")
    def UserNotValidated(self, app_name, result):
        """Signal thrown when the user is not validated."""

    @dbus.service.method(dbus_interface=DBUS_IFACE_USER_NAME,
                         in_signature='sss')
    def login(self, app_name, email, password):
        """Call the matching method in the processor."""
        self.root.sso_login.login(app_name, email, password, ping_url=None)

    @dbus.service.method(dbus_interface=DBUS_IFACE_USER_NAME,
                         in_signature='ssss')
    def login_and_ping(self, app_name, email, password, ping_url):
        """Call the matching method in the processor."""
        self.root.sso_login.login(app_name, email, password, ping_url)

    # validate_email signals
    @dbus.service.signal(DBUS_IFACE_USER_NAME, signature="ss")
    def EmailValidated(self, app_name, result):
        """Signal thrown after the email is validated."""

    @dbus.service.signal(DBUS_IFACE_USER_NAME, signature="sa{ss}")
    def EmailValidationError(self, app_name, error):
        """Signal thrown when there's a problem validating the email."""

    @dbus.service.method(dbus_interface=DBUS_IFACE_USER_NAME,
                         in_signature='ssss')
    def validate_email(self, app_name, email, password, email_token):
        """Call the matching method in the processor."""
        self.root.sso_login.validate_email(app_name,
            email, password, email_token, ping_url=None)

    @dbus.service.method(dbus_interface=DBUS_IFACE_USER_NAME,
                         in_signature='sssss')
    def validate_email_and_ping(self, app_name, email, password, email_token,
                                ping_url):
        """Call the matching method in the processor."""
        self.root.sso_login.validate_email(app_name,
            email, password, email_token, ping_url)

    # request_password_reset_token signals
    @dbus.service.signal(DBUS_IFACE_USER_NAME, signature="ss")
    def PasswordResetTokenSent(self, app_name, result):
        """Signal thrown when the token is succesfully sent."""

    @dbus.service.signal(DBUS_IFACE_USER_NAME, signature="sa{ss}")
    def PasswordResetError(self, app_name, error):
        """Signal thrown when there's a problem sending the token."""

    @dbus.service.method(dbus_interface=DBUS_IFACE_USER_NAME,
                         in_signature='ss')
    def request_password_reset_token(self, app_name, email):
        """Call the matching method in the processor."""
        self.root.sso_login.request_password_reset_token(app_name, email)

    # set_new_password signals
    @dbus.service.signal(DBUS_IFACE_USER_NAME, signature="ss")
    def PasswordChanged(self, app_name, result):
        """Signal thrown when the token is succesfully sent."""

    @dbus.service.signal(DBUS_IFACE_USER_NAME, signature="sa{ss}")
    def PasswordChangeError(self, app_name, error):
        """Signal thrown when there's a problem sending the token."""

    @dbus.service.method(dbus_interface=DBUS_IFACE_USER_NAME,
                         in_signature='ssss')
    def set_new_password(self, app_name, email, token, new_password):
        """Call the matching method in the processor."""
        self.root.sso_login.set_new_password(app_name,
            email, token, new_password)


class CredentialsManagementProxy(dbus.service.Object):
    """Object that manages credentials.

    Every exposed method in this class requires one mandatory argument:

        - 'app_name': the name of the application. Will be displayed in the
        GUI header, plus it will be used to find/build/clear tokens.

    And accepts another parameter named 'args', which is a dictionary that
    can contain the following:

        - 'help_text': an explanatory text for the end-users, will be
        shown below the header. This is an optional free text field.

        - 'ping_url': the url to open after successful token retrieval. If
        defined, the email will be attached to the url and will be pinged
        with a OAuth-signed request.

        - 'tc_url': the link to the Terms and Conditions page. If defined,
        the checkbox to agree to the terms will link to it.

        - 'window_id': the id of the window which will be set as a parent
        of the GUI. If not defined, no parent will be set.

    """

    # Use of super on an old style class
    # pylint: disable=E1002

    def __init__(self, root, *args, **kwargs):
        # pylint: disable=E1002
        super(CredentialsManagementProxy, self).__init__(*args, **kwargs)
        self.root = root

    # Operator not preceded by a space (fails with dbus decorators)
    # pylint: disable=C0322

    @dbus.service.signal(DBUS_CREDENTIALS_IFACE, signature='s')
    def AuthorizationDenied(self, app_name):
        """Signal thrown when the user denies the authorization."""

    @dbus.service.signal(DBUS_CREDENTIALS_IFACE, signature='sa{ss}')
    def CredentialsFound(self, app_name, credentials):
        """Signal thrown when the credentials are found."""

    @dbus.service.signal(DBUS_CREDENTIALS_IFACE, signature='s')
    def CredentialsNotFound(self, app_name):
        """Signal thrown when the credentials are not found."""

    @dbus.service.signal(DBUS_CREDENTIALS_IFACE, signature='s')
    def CredentialsCleared(self, app_name):
        """Signal thrown when the credentials were cleared."""

    @dbus.service.signal(DBUS_CREDENTIALS_IFACE, signature='s')
    def CredentialsStored(self, app_name):
        """Signal thrown when the credentials were cleared."""

    @dbus.service.signal(DBUS_CREDENTIALS_IFACE, signature='sa{ss}')
    def CredentialsError(self, app_name, error_dict):
        """Signal thrown when there is a problem getting the credentials."""

    @dbus.service.method(dbus_interface=DBUS_CREDENTIALS_IFACE,
                         in_signature='sa{ss}', out_signature='')
    def find_credentials(self, app_name, args):
        """Look for the credentials for an application.

        - 'app_name': the name of the application which credentials are
        going to be removed.

        - 'args' is a dictionary, currently not used.

        """
        self.root.cred_manager.find_credentials(app_name, args)

    @dbus.service.method(dbus_interface=DBUS_CREDENTIALS_IFACE,
                         in_signature="sa{ss}", out_signature="a{ss}",
                         async_callbacks=("reply_handler", "error_handler"))
    def find_credentials_sync(self, app_name, args,
                              reply_handler=NO_OP, error_handler=NO_OP):
        """Get the credentials from the keyring or {} if not there.

        This method SHOULD NOT be used, is here only for compatibilty issues.

        """

        def _drop_dict(error_dict):
            """Call 'error_handler' properly."""
            error_handler(dbus.service.DBusException(error_dict['errtype']))

        self.root.cred_manager.find_credentials(app_name, args,
                                   success_cb=reply_handler,
                                   error_cb=_drop_dict)

    @dbus.service.method(dbus_interface=DBUS_CREDENTIALS_IFACE,
                         in_signature='sa{ss}', out_signature='')
    def clear_credentials(self, app_name, args):
        """Clear the credentials for an application.

        - 'app_name': the name of the application which credentials are
        going to be removed.

        - 'args' is a dictionary, currently not used.

        """
        self.root.cred_manager.clear_credentials(app_name, args)

    @dbus.service.method(dbus_interface=DBUS_CREDENTIALS_IFACE,
                         in_signature='sa{ss}', out_signature='')
    def store_credentials(self, app_name, args):
        """Store the token for an application.

        - 'app_name': the name of the application which credentials are
        going to be stored.

        - 'args' is the dictionary holding the credentials. Needs to provide
        the following mandatory keys: 'token', 'token_key', 'consumer_key',
        'consumer_secret'.

        """
        self.root.cred_manager.store_credentials(app_name, args)

    @dbus.service.method(dbus_interface=DBUS_CREDENTIALS_IFACE,
                         in_signature='sa{ss}', out_signature='')
    def register(self, app_name, args):
        """Get credentials if found else prompt GUI to register."""
        self.root.cred_manager.register(app_name, args)

    @dbus.service.method(dbus_interface=DBUS_CREDENTIALS_IFACE,
                         in_signature='sa{ss}', out_signature='')
    def login(self, app_name, args):
        """Get credentials if found else prompt GUI to login."""
        self.root.cred_manager.login(app_name, args)

    @dbus.service.method(dbus_interface=DBUS_CREDENTIALS_IFACE,
                         in_signature='sa{ss}', out_signature='')
    def login_email_password(self, app_name, args):
        """Get credentials if found, else login using email and password.

        - 'args' should contain at least the follwing keys: 'email' and
        'password'. Those will be used to issue a new SSO token, which will be
        returned trough the CredentialsFound signal.

        """
        self.root.cred_manager.login_email_password(app_name, args)


class UbuntuSSOProxy(object):
    """Object that exposes the diff referenceable objects."""

    def __init__(self, root):
        self.root = root
        self.sso_login = None
        self.cred_manager = None

        try:
            self.bus = dbus.SessionBus()
        except dbus.service.DBusException as e:
            logger.exception(e)
            shutdown_func()

    def start(self):
        """Start listening, nothing async to be done in this platform."""
        # Register DBus service for making sure we run only one instance
        name = self.bus.request_name(DBUS_BUS_NAME,
                                     dbus.bus.NAME_FLAG_DO_NOT_QUEUE)
        if name == dbus.bus.REQUEST_NAME_REPLY_EXISTS:
            raise AlreadyStartedError()

        bus_name = dbus.service.BusName(DBUS_BUS_NAME, bus=self.bus)
        self.sso_login = SSOLoginProxy(self.root,
                                       bus_name=bus_name,
                                       object_path=DBUS_ACCOUNT_PATH)
        self.cred_manager = CredentialsManagementProxy(self.root,
                                bus_name=bus_name,
                                object_path=DBUS_CREDENTIALS_PATH)

        return defer.succeed(None)

    def shutdown(self):
        """Shutdown the service."""
        self.sso_login.remove_from_connection()
        self.cred_manager.remove_from_connection()
        self.bus.release_name(DBUS_BUS_NAME)
        return defer.succeed(None)


# ============================== client classes ==============================


class RemoteClient(object):
    """Client that can perform calls to remote DBus object."""

    bus_name = None
    path = None
    interface = None

    def __init__(self):
        self.bus = dbus.SessionBus()
        obj = self.bus.get_object(bus_name=self.bus_name,
                                  object_path=self.path,
                                  follow_name_owner_changes=True)
        self.dbus_iface = dbus.Interface(obj, dbus_interface=self.interface)
        self.dbus_iface.call_method = self.call_method
        self.dbus_iface.disconnect_from_signal = lambda _, sig: sig.remove()

    def call_method(self, method_name, *args, **kwargs):
        """Call asynchronously 'method_name(*args)'.

        Return a deferred that will be fired when the call finishes.

        """
        d = defer.Deferred()

        reply_handler = kwargs.get('reply_handler', None)
        if reply_handler is not None:
            d.addCallback(lambda a: reply_handler(*a))

        error_handler = kwargs.get('error_handler', None)
        if error_handler is not None:
            d.addErrback(lambda f: error_handler(f.value))

        self.bus.call_async(
            bus_name=self.bus_name, object_path=self.path,
            dbus_interface=self.interface, method=method_name,
            signature=None, args=args,
            reply_handler=lambda *a: d.callback(a),
            error_handler=d.errback)

        return d


class SSOLoginClient(RemoteClient):
    """Access the UserManagement DBus interface."""

    bus_name = DBUS_BUS_NAME
    path = DBUS_ACCOUNT_PATH
    interface = DBUS_IFACE_USER_NAME


class CredentialsManagementClient(RemoteClient):
    """Access the CredentialsManagement DBus interface."""

    bus_name = DBUS_BUS_NAME
    path = DBUS_CREDENTIALS_PATH
    interface = DBUS_CREDENTIALS_IFACE


class UbuntuSSOClient(object):
    """Base client that provides remote access to the sso API."""

    def __init__(self):
        self.sso_login = SSOLoginClient().dbus_iface
        self.cred_manager = CredentialsManagementClient().dbus_iface

    def connect(self):
        """No need to connect DBus proxy objects."""
        return defer.succeed(None)

    def disconnect(self):
        """No need to disconnect DBus proxy objects."""
        return defer.succeed(None)


def get_sso_client():
    """Get a client to access the SSO service."""
    result = UbuntuSSOClient()
    return defer.succeed(result)


try:
    from ubuntu_kylin_sso.main import qt
    source = qt
except ImportError:  # no PyQt4.QtCore in the system
    from ubuntu_kylin_sso.main import glib
    source = glib


timeout_func = source.timeout_func
shutdown_func = source.shutdown_func
start_setup = source.start_setup


def sighup_handler(*a, **kw):
    """Stop the service."""
    # This handler may be called in any thread, so is not thread safe.
    # See the link below for info:
    # www.listware.net/201004/gtk-devel-list/115067-unix-signals-in-glib.html
    logger.info("Stoping Ubuntu SSO service since SIGHUP was received.")
    shutdown_func()


class AlreadyStartedError(Exception):
    """The backend service has already been started."""


def finish_setup(result, loop):
    """Run the specific mainloop only if no failure ocurred."""
    if result is None:  # no failure ocurred, start the service
        logger.debug("Hooking up SIGHUP with handler %r.", sighup_handler)
        signal.signal(signal.SIGHUP, sighup_handler)
        source.run_func(loop)


def main():
    """Run the specific mainloop."""



# -*- coding: utf-8 -*-
#
# Copyright 2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Darwin specific tests for the main module."""

import os

from dirspec import basedir
from twisted.internet import defer

from ubuntu_kylin_sso.main import darwin
from ubuntu_kylin_sso.tests import TestCase

# because we are using twisted we have java like names C0103
# pylint: disable=C0103


class DarwinTestCase(TestCase):
    """Tests for module level misc functions"""

    def test_get_sso_pb_port(self):
        """Test the get_sso_pb_port function, by patching os.getuid"""
        expected = os.path.join(basedir.xdg_cache_home, 'sso', 'ipc')
        self.assertEqual(darwin.get_sso_domain_socket(), expected)


class DescriptionFactoryTestcase(TestCase):
    """Test the factory."""

    @defer.inlineCallbacks
    def setUp(self):
        """Set the tests."""
        yield super(DescriptionFactoryTestcase, self).setUp()
        self.domain = '/path/to/domain'
        self.patch(darwin, 'get_sso_domain_socket', lambda: self.domain)

    def test_server_description(self):
        """Test getting the description info."""
        expected = darwin.DescriptionFactory.server_description_pattern % \
                       self.domain
        factory = darwin.DescriptionFactory()
        self.assertEqual(expected, factory.server)

    def test_client_description(self):
        """Test getting the description info."""
        expected = darwin.DescriptionFactory.client_description_pattern % \
                       self.domain
        factory = darwin.DescriptionFactory()
        self.assertEqual(expected, factory.client)
